from cumulusci.tasks.github import (
    CloneTag,
    CreateRelease,
    GetPackageDataFromCommitStatus,
    MergeBranch,
    PublishSubtree,
    PullRequests,
    ReleaseReport,
)


class TestVCSMigration:
    def test_clone_tag(self):
        assert CloneTag.__module__ == "cumulusci.tasks.github.tag"
        assert CloneTag.__name__ == "CloneTag"

    def test_merge_branch(self):
        assert MergeBranch.__module__ == "cumulusci.tasks.github.merge"
        assert MergeBranch.__name__ == "MergeBranch"

    def test_get_package_data_from_commit_status(self):
        assert (
            GetPackageDataFromCommitStatus.__module__
            == "cumulusci.tasks.github.commit_status"
        )
        assert (
            GetPackageDataFromCommitStatus.__name__ == "GetPackageDataFromCommitStatus"
        )

    def test_publish_subtree(self):
        assert PublishSubtree.__module__ == "cumulusci.tasks.github.publish"
        assert PublishSubtree.__name__ == "PublishSubtree"

    def test_pull_requests(self):
        assert PullRequests.__module__ == "cumulusci.tasks.github.pull_request"
        assert PullRequests.__name__ == "PullRequests"

    def test_release(self):
        assert CreateRelease.__module__ == "cumulusci.tasks.github.release"
        assert CreateRelease.__name__ == "CreateRelease"

    def test_release_report(self):
        assert ReleaseReport.__module__ == "cumulusci.tasks.github.release_report"
        assert ReleaseReport.__name__ == "ReleaseReport"



import enum


class StrEnum(str, enum.Enum):
    """Shim to preserve pre-Python 3.11 behavior
    for the string values of enum members.

    This can be replaced by enum.StrEnum when 3.11
    is the oldest supported version."""

    __str__ = str.__str__  # type: ignore



import http.client
import os
import shutil
import tempfile
from datetime import datetime, timedelta
from unittest import mock

import pytest
import responses

from cumulusci.core.exceptions import GithubApiNotFoundError
from cumulusci.core.github import get_github_api
from cumulusci.tasks.github.tests.util_github_api import GithubApiTestMixin
from cumulusci.tasks.release_notes.provider import (
    BaseChangeNotesProvider,
    DirectoryChangeNotesProvider,
    StaticChangeNotesProvider,
)
from cumulusci.tasks.release_notes.tests.utils import MockUtil
from cumulusci.vcs.github.release_notes.generator import GithubReleaseNotesGenerator
from cumulusci.vcs.github.release_notes.provider import GithubChangeNotesProvider

__location__ = os.path.split(os.path.realpath(__file__))[0]
date_format = "%Y-%m-%dT%H:%M:%SZ"
PARSER_CONFIG = [
    {
        "class_path": "cumulusci.vcs.github.release_notes.parser.GithubLinesParser",
        "title": "Critical Changes",
    },
    {
        "class_path": "cumulusci.vcs.github.release_notes.parser.GithubLinesParser",
        "title": "Changes",
    },
    {
        "class_path": "cumulusci.vcs.github.release_notes.parser.GithubIssuesParser",
        "title": "Issues Closed",
    },
]


class TestBaseChangeNotesProvider:
    def test_init(self):
        provider = BaseChangeNotesProvider("test")
        assert provider.release_notes_generator == "test"

    def test_call_raises_notimplemented(self):
        provider = BaseChangeNotesProvider("test")
        with pytest.raises(NotImplementedError):
            provider.__call__()


class TestStaticChangeNotesProvider:
    def test_empty_list(self):
        provider = StaticChangeNotesProvider("test", [])
        assert list(provider()) == []

    def test_single_item_list(self):
        provider = StaticChangeNotesProvider("test", ["abc"])
        assert list(provider()) == ["abc"]

    def test_multi_item_list(self):
        provider = StaticChangeNotesProvider("test", ["abc", "d", "e"])
        assert list(provider()) == ["abc", "d", "e"]


class TestDirectoryChangeNotesProvider:
    def get_empty_dir(self):
        tempdir = tempfile.mkdtemp()
        return os.path.join(tempdir)

    def get_dir_content(self, path):
        dir_content = []
        for item in sorted(os.listdir(path)):
            item_path = "{}/{}".format(path, item)
            dir_content.append(open(item_path, "r").read())
        return dir_content

    def test_empty_directory(self):
        directory = self.get_empty_dir()
        provider = DirectoryChangeNotesProvider("test", directory)
        dir_content = self.get_dir_content(directory)
        assert list(provider()) == dir_content
        shutil.rmtree(directory)

    def test_single_item_directory(self):
        directory = "{}/change_notes/single/".format(__location__)
        provider = DirectoryChangeNotesProvider("test", directory)
        dir_content = self.get_dir_content(directory)
        assert list(provider()) == dir_content

    def test_multi_item_directory(self):
        directory = "{}/change_notes/multi/".format(__location__)
        provider = DirectoryChangeNotesProvider("test", directory)
        dir_content = self.get_dir_content(directory)
        assert list(provider()) == dir_content


class TestGithubChangeNotesProvider(GithubApiTestMixin):
    def setup_method(self):
        # Set up the mock release_tag lookup response
        self.repo_api_url = "https://api.github.com/repos/TestOwner/TestRepo"
        # Tag that does not exist
        self.invalid_tag = "release/1.4"
        # The current production release
        self.current_tag = "release/1.3"
        # The previous beta release
        self.beta_tag = "beta/1.3-Beta_1"
        # The previous production release with no change notes vs 1.3
        self.last_tag = "release/1.2"
        # The second previous production release with one change note vs 1.3
        self.last2_tag = "release/1.1"
        # The third previous production release with three change notes vs 1.3
        self.last3_tag = "release/1.0"

        self.current_tag_sha = self._random_sha()
        self.beta_tag_sha = self._random_sha()
        self.current_tag_commit_sha = self._random_sha()
        self.current_tag_commit_date = datetime.utcnow()
        self.last_tag_sha = self._random_sha()
        self.last_tag_commit_sha = self._random_sha()
        self.last_tag_commit_date = datetime.utcnow() - timedelta(days=1)
        self.last2_tag_sha = self._random_sha()
        self.gh = get_github_api("TestUser", "TestPass")
        self.init_github()
        self.mock_util = MockUtil("TestOwner", "TestRepo")

    def _create_generator(self, current_tag, last_tag=None):
        generator = GithubReleaseNotesGenerator(
            self.gh,
            self.github_info.copy(),
            PARSER_CONFIG,
            current_tag,
            last_tag=last_tag,
        )
        return generator

    def _mock_current_tag(self):
        api_url = "{}/git/tags/{}".format(self.repo_api_url, self.current_tag_sha)
        expected_response = self._get_expected_tag(
            self.current_tag,
            self.current_tag_commit_sha,
            self.current_tag_sha,
            self.current_tag_commit_date,
        )
        responses.add(method=responses.GET, url=api_url, json=expected_response)
        return expected_response

    def _mock_current_tag_commit(self):
        api_url = "{}/git/commits/{}".format(
            self.repo_api_url, self.current_tag_commit_sha
        )
        expected_response = {
            "author": {
                "name": "John Doe",
                "email": "john.doe@example.com",
                "date": datetime.strftime(self.current_tag_commit_date, date_format),
            },
            "committer": None,
            "message": "",
            "parents": [],
            "sha": self.current_tag_commit_sha,
            "tree": {"sha": "", "url": ""},
            "url": "",
            "verification": None,
        }
        responses.add(method=responses.GET, url=api_url, json=expected_response)

    def _mock_current_tag_ref(self):
        api_url = "{}/git/ref/tags/{}".format(self.repo_api_url, self.current_tag)
        expected_response_current_tag_ref = self._get_expected_tag_ref(
            self.current_tag, self.current_tag_sha
        )
        responses.add(
            method=responses.GET, url=api_url, json=expected_response_current_tag_ref
        )

    def _mock_invalid_tag(self):
        api_url = "{}/git/ref/tags/{}".format(self.repo_api_url, self.invalid_tag)
        expected_response = {
            "message": "Not Found",
            "documentation_url": "https://developer.github.com/v3",
        }
        responses.add(
            method=responses.GET,
            url=api_url,
            json=expected_response,
            status=http.client.NOT_FOUND,
        )

    def _mock_last_tag(self):
        api_url = "{}/git/tags/{}".format(self.repo_api_url, self.last_tag_sha)
        expected_response = self._get_expected_tag(
            self.last_tag,
            self.last_tag_commit_sha,
            self.last_tag_sha,
            self.last_tag_commit_date,
        )
        responses.add(method=responses.GET, url=api_url, json=expected_response)
        return expected_response

    def _mock_last_tag_commit(self):
        api_url = "{}/git/commits/{}".format(
            self.repo_api_url, self.last_tag_commit_sha
        )
        expected_response = {
            "author": {
                "name": "John Doe",
                "date": datetime.strftime(self.last_tag_commit_date, date_format),
            },
            "committer": None,
            "message": "",
            "parents": [],
            "sha": self.last_tag_commit_sha,
            "tree": {"sha": "", "url": ""},
            "url": "",
            "verification": None,
        }
        responses.add(method=responses.GET, url=api_url, json=expected_response)

    def _mock_last_tag_ref(self):
        api_url = "{}/git/ref/tags/{}".format(self.repo_api_url, self.last_tag)
        expected_response_last_tag_ref = self._get_expected_tag_ref(
            self.last_tag, self.last_tag_sha
        )
        responses.add(
            method=responses.GET, url=api_url, json=expected_response_last_tag_ref
        )

    def _mock_list_pull_requests_one_in_range(self):
        api_url = "{}/pulls".format(self.repo_api_url)
        expected_response = [
            self._get_expected_pull_request(
                1, 101, "pull 1", datetime.utcnow() - timedelta(seconds=60)
            ),
            self._get_expected_pull_request(
                2, 102, "pull 2", datetime.utcnow() - timedelta(days=4)
            ),
            self._get_expected_pull_request(
                3, 103, "pull 3", datetime.utcnow() - timedelta(days=5)
            ),
        ]
        responses.add(method=responses.GET, url=api_url, json=expected_response)

    def _mock_list_pull_requests_multiple_in_range(self):
        api_url = "{}/pulls".format(self.repo_api_url)
        expected_response = [
            self._get_expected_pull_request(
                1, 101, "pull 1", datetime.utcnow() - timedelta(seconds=60)
            ),
            self._get_expected_pull_request(
                2, 102, "pull 2", datetime.utcnow() - timedelta(seconds=90)
            ),
            self._get_expected_pull_request(
                3, 103, "pull 3", datetime.utcnow() - timedelta(seconds=120)
            ),
            self._get_expected_pull_request(
                4, 104, "pull 4", datetime.utcnow() - timedelta(days=4)
            ),
            self._get_expected_pull_request(
                5, 105, "pull 5", datetime.utcnow() - timedelta(days=5)
            ),
            self._get_expected_pull_request(6, 106, "pull 6", None),
            self._get_expected_pull_request(
                7,
                107,
                "pull 7",
                datetime.utcnow() - timedelta(seconds=180),
                merge_commit_sha=self.last_tag_commit_sha,
            ),
            self._get_expected_pull_request(
                8,
                108,
                "pull 8",
                datetime.utcnow(),
                merge_commit_sha=self.current_tag_commit_sha,
            ),
        ]
        responses.add(method=responses.GET, url=api_url, json=expected_response)

    def _mock_list_tags_multiple(self):
        api_url = "{}/tags".format(self.repo_api_url)
        expected_response = [
            self._get_expected_repo_tag(self.current_tag, self.current_tag_sha),
            self._get_expected_repo_tag(self.beta_tag, self.beta_tag_sha),
            self._get_expected_repo_tag(self.last_tag, self.last_tag_sha),
            self._get_expected_repo_tag(self.last2_tag, self.last2_tag_sha),
        ]
        responses.add(method=responses.GET, url=api_url, json=expected_response)

    def _mock_list_tags_single(self):
        api_url = "{}/tags".format(self.repo_api_url)
        expected_response = [
            self._get_expected_repo_tag(self.current_tag, self.current_tag_sha)
        ]
        responses.add(method=responses.GET, url=api_url, json=expected_response)

    @responses.activate
    def test_invalid_current_tag(self):
        self.mock_util.mock_get_repo()
        self._mock_invalid_tag()
        generator = self._create_generator(self.invalid_tag)
        provider = GithubChangeNotesProvider(generator, self.invalid_tag)
        with pytest.raises(GithubApiNotFoundError):
            provider.current_tag_info

    @responses.activate
    def test_current_tag_without_last(self):
        self.mock_util.mock_get_repo()
        self._mock_current_tag_ref()
        expected_current_tag = self._mock_current_tag()
        self._mock_current_tag_commit()
        self._mock_last_tag_ref()
        expected_last_tag = self._mock_last_tag()
        self._mock_last_tag_commit()
        self._mock_list_tags_multiple()

        generator = self._create_generator(self.current_tag)
        provider = GithubChangeNotesProvider(generator, self.current_tag)
        current_tag = provider.current_tag_info["tag"]
        last_tag = provider.last_tag_info["tag"]

        assert current_tag.tag == expected_current_tag["tag"]
        assert last_tag.tag == expected_last_tag["tag"]

    @responses.activate
    def test_current_tag_without_last_no_last_found(self):
        self.mock_util.mock_get_repo()
        self._mock_current_tag_ref()
        self._mock_current_tag()
        self._mock_current_tag_commit()
        self._mock_list_tags_single()

        generator = self._create_generator(self.current_tag)
        provider = GithubChangeNotesProvider(generator, self.current_tag)

        assert provider.last_tag is None
        assert provider.last_tag_info is None

    @responses.activate
    def test_no_pull_requests_in_repo(self):
        self.mock_util.mock_get_repo()
        # Mock the tag calls
        self._mock_current_tag_ref()
        self._mock_current_tag()
        self._mock_current_tag_commit()
        self._mock_last_tag_ref()
        self._mock_last_tag()
        self._mock_last_tag_commit()

        # Mock the list all pull requests call
        api_url = "{}/pulls".format(self.repo_api_url)
        responses.add(
            method=responses.GET, url=api_url, json=[], content_type="application/json"
        )

        generator = self._create_generator(self.current_tag, self.last_tag)
        provider = GithubChangeNotesProvider(generator, self.current_tag, self.last_tag)
        assert list(provider()) == []

    @responses.activate
    def test_no_pull_requests_in_range(self):
        self.mock_util.mock_get_repo()
        # Mock the tag calls
        self._mock_current_tag_ref()
        self._mock_current_tag()
        self._mock_current_tag_commit()
        self._mock_last_tag_ref()
        self._mock_last_tag()
        self._mock_last_tag_commit()

        # Mock the list all pull requests call
        api_url = "{}/pulls".format(self.repo_api_url)
        expected_pull_request_1 = self._get_expected_pull_request(
            pull_id=1,
            issue_number=101,
            body="pull 1",
            merged_date=datetime.utcnow() - timedelta(days=2),
        )
        expected_response_list_pull_requests = [expected_pull_request_1]
        responses.add(
            method=responses.GET, url=api_url, json=expected_response_list_pull_requests
        )

        generator = self._create_generator(self.current_tag, self.last_tag)
        provider = GithubChangeNotesProvider(generator, self.current_tag, self.last_tag)
        assert list(provider()) == []

    @responses.activate
    def test_one_pull_request_in_range(self):
        self.mock_util.mock_get_repo()
        # Mock the tag calls
        self._mock_current_tag_ref()
        self._mock_current_tag()
        self._mock_current_tag_commit()
        self._mock_last_tag_ref()
        self._mock_last_tag()
        self._mock_last_tag_commit()
        self._mock_list_pull_requests_one_in_range()

        generator = self._create_generator(self.current_tag, self.last_tag)
        provider = GithubChangeNotesProvider(generator, self.current_tag, self.last_tag)
        provider_list = list(provider())
        pr_body_list = ["pull 1"]
        assert len(provider_list) == len(pr_body_list)
        for pr, pr_body in zip(provider_list, pr_body_list):
            assert pr.body == pr_body

    @responses.activate
    def test_multiple_pull_requests_in_range(self):
        self.mock_util.mock_get_repo()
        # Mock the tag calls
        self._mock_current_tag_ref()
        self._mock_current_tag()
        self._mock_current_tag_commit()
        self._mock_last_tag_ref()
        self._mock_last_tag()
        self._mock_last_tag_commit()
        self._mock_list_pull_requests_multiple_in_range()

        generator = self._create_generator(self.current_tag, self.last_tag)
        provider = GithubChangeNotesProvider(generator, self.current_tag, self.last_tag)
        provider_list = list(provider())
        pr_body_list = []
        pr_body_list = ["pull 1", "pull 2", "pull 3", "pull 8"]
        assert len(provider_list) == len(pr_body_list)
        for pr, pr_body in zip(provider_list, pr_body_list):
            assert pr.body == pr_body

    @responses.activate
    def test_pull_requests_with_no_last_tag(self):
        self.mock_util.mock_get_repo()
        # Mock the tag calls
        self._mock_current_tag_ref()
        self._mock_current_tag()
        self._mock_current_tag_commit()
        self._mock_last_tag_ref()
        self._mock_last_tag()
        self._mock_last_tag_commit()
        self._mock_list_pull_requests_multiple_in_range()

        generator = self._create_generator(self.current_tag)
        provider = GithubChangeNotesProvider(generator, self.current_tag)
        provider._get_last_tag = mock.Mock(return_value=None)
        provider_list = list(provider())
        pr_body_list = []
        pr_body_list = [
            "pull 1",
            "pull 2",
            "pull 3",
            "pull 4",
            "pull 5",
            "pull 7",
            "pull 8",
        ]
        assert len(provider_list) == len(pr_body_list)
        for pr, pr_body in zip(provider_list, pr_body_list):
            assert pr.body == pr_body

    @responses.activate
    def test_get_version_from_tag(self):
        self.mock_util.mock_get_repo()
        tag = "beta/1.0-Beta_1"
        generator = self._create_generator(tag)
        provider = GithubChangeNotesProvider(generator, tag)
        assert provider._get_version_from_tag(tag) == "1.0-Beta_1"
        with pytest.raises(ValueError):
            provider._get_version_from_tag("bogus")


class TestMigrationParser(GithubApiTestMixin):
    def setup_method(self):
        self.init_github()
        self.gh = get_github_api("TestUser", "TestPass")
        self.title = "Issues"
        # Set up the mock release_tag lookup response
        self.issue_number_valid = 123
        self.issue_number_invalid = 456
        self.pr_number = 789
        self.pr_url = "https://github.com/{}/{}/pulls/{}".format(
            "TestOwner", "TestRepo", self.pr_number
        )
        self.mock_util = MockUtil("TestOwner", "TestRepo")

    @responses.activate
    def test_migration(self):
        from cumulusci.tasks.release_notes.provider import GithubChangeNotesProvider
        from cumulusci.utils.deprecation import ClassMovedWarning

        self.mock_util.mock_get_repo()
        tag = "beta/1.0-Beta_1"
        generator = self._create_generator(tag)

        with pytest.warns(
            ClassMovedWarning, match="cumulusci.vcs.github.release_notes.provider"
        ):
            GithubChangeNotesProvider(generator, tag)

        assert (
            GithubChangeNotesProvider.__module__
            == "cumulusci.tasks.release_notes.provider"
        )
        assert GithubChangeNotesProvider.__name__ == "GithubChangeNotesProvider"

    def _create_generator(self, current_tag, last_tag=None):
        generator = GithubReleaseNotesGenerator(
            self.gh,
            self.github_info.copy(),
            PARSER_CONFIG,
            current_tag,
            last_tag=last_tag,
        )
        return generator



from cumulusci.core.config import BaseProjectConfig
from cumulusci.core.config.sfdx_org_config import SfdxOrgConfig
from cumulusci.core.exceptions import CumulusCIException, ServiceNotConfigured
from cumulusci.core.sfdx import get_default_devhub_username


def get_devhub_config(project_config: BaseProjectConfig) -> SfdxOrgConfig:
    """
    @param project_config: a base project configuration
    @return: an SfdxOrgConfig tied to the devHub
    """
    try:
        devhub_service = project_config.keychain.get_service("devhub")
    except (ServiceNotConfigured, CumulusCIException):
        devhub_username = get_default_devhub_username()
    else:
        devhub_username = devhub_service.username
    return SfdxOrgConfig({"username": devhub_username}, "devhub")



from typing import Optional


def format_subscriber_package_version_where_clause(
    spv_id: str, install_key: Optional[str] = None
) -> str:
    """Get the where clause for a SubscriberPackageVersion query

    Does not include the WHERE.
    Includes the installation key if provided.
    """
    where_clause = f"Id='{spv_id}'"

    if install_key:
        where_clause += f" AND InstallationKey ='{install_key}'"

    return where_clause



import json
import os
import shutil
import tempfile
from collections import defaultdict
from itertools import chain
from xml.etree.ElementTree import ParseError

from defusedxml.minidom import parseString

from cumulusci.core.config import FlowConfig, TaskConfig
from cumulusci.core.exceptions import TaskOptionsError
from cumulusci.core.flowrunner import FlowCoordinator
from cumulusci.core.sfdx import convert_sfdx_source
from cumulusci.core.utils import process_list_arg
from cumulusci.salesforce_api.metadata import ApiRetrieveUnpackaged
from cumulusci.tasks.metadata.package import process_common_components
from cumulusci.tasks.salesforce import BaseSalesforceTask
from cumulusci.utils import cd
from cumulusci.utils.xml import metadata_tree


class CheckComponents(BaseSalesforceTask):
    api_retrieve_unpackaged = ApiRetrieveUnpackaged
    task_options = {
        "paths": {
            "description": "List of deploy paths to check",
            "required": False,
        },
        "name": {
            "description": "The name of the current plan or flow to detect deploy paths",
            "required": False,
        },
    }
    deploy_paths = []

    def _init_options(self, kwargs):
        super(CheckComponents, self)._init_options(kwargs)
        if "paths" in self.options and "name" in self.options:
            raise TaskOptionsError("Please provide either --paths or --name")
        if "paths" not in self.options and "name" not in self.options:
            raise TaskOptionsError(
                "This task requires a plan/flow name or paths options. Pass --paths or --name options"
            )

    def _run_task(self):
        # Check if paths are provided in options. Assuming to only check for those paths
        paths = self.options.get("paths")
        plan_or_flow_name = self.options.get("name")

        (
            components,
            api_retrieve_unpackaged_response,
        ) = self.get_repo_existing_components(plan_or_flow_name, paths)

        if not components:
            self.logger.info("No components found in deploy path")
            raise TaskOptionsError("No plan or paths options provided")

        self.logger.debug("Components detected at source")
        for component_type, component_names in components.items():
            self.logger.debug(f"{component_type}: {', '.join(component_names)}")
        # check common components
        components.pop("Settings", None)
        existing_components = process_common_components(
            api_retrieve_unpackaged_response, components
        )

        if existing_components:
            self.logger.info("Components exists in the target org:")
            for component_type, component_names in existing_components.items():
                self.logger.info(f"{component_type}: {', '.join(component_names)}")
            self.return_values["existing_components"] = existing_components
        else:
            self.logger.info(
                "No components from the deploy paths exist in the target org."
            )

    def get_repo_existing_components(self, plan_or_flow_name, paths=""):
        if paths:
            paths = process_list_arg(paths)
            self.logger.info(f"Using provided paths: {paths}")
            self.deploy_paths = paths
        elif plan_or_flow_name:
            # if path is not provided
            is_plan = self._is_plan(plan_or_flow_name)
            if is_plan is None:
                raise TaskOptionsError(
                    f"Plan or flow name '{plan_or_flow_name}' not found"
                )

            self.logger.info(
                f"Analyzing project {'plan' if is_plan else 'flow'}: {plan_or_flow_name}"
            )

            # load deploy paths from all the steps in plan or flow
            self._load_deploy_paths(plan_or_flow_name, is_plan)
            if not self.deploy_paths:
                self.logger.warning("No deploy paths found in the plan or flow.")
                return
            self.logger.debug(
                f"deploy paths found in the plan or flow.{self.deploy_paths}"
            )
        # Temp dir to copy all deploy paths from task options
        temp_dir = tempfile.mkdtemp()
        self.logger.info(f"Temporary deploy directory created: {temp_dir}")
        mdapi_components = {}
        mdapi_response_messages = []
        for path in self.deploy_paths:
            full_path = os.path.join(self.project_config.repo_root, path)
            if not os.path.exists(full_path):
                self.logger.info(f"Skipping path: '{path}' - path doesn't exist")
                continue
            elif "package.xml" in os.listdir(full_path):
                package_xml_path = os.path.join(full_path, "package.xml")
                source_xml_tree = metadata_tree.parse(package_xml_path)
                components = metadata_tree.parse_package_xml_types(
                    "name", source_xml_tree
                )
                response_messages = self._get_api_object_responce(
                    package_xml_path, source_xml_tree.version.text
                )
                merged = {}
                for key in set(components).union(mdapi_components):
                    merged[key] = list(
                        set(
                            chain(
                                components.get(key, []), mdapi_components.get(key, [])
                            )
                        )
                    )
                mdapi_components = merged
                mdapi_response_messages.extend(response_messages)
                continue
            self._copy_to_tempdir(path, temp_dir)

        (
            components,
            api_retrieve_unpackaged_response,
        ) = self._collect_components_from_paths(temp_dir)

        # remove temp dir
        shutil.rmtree(temp_dir)
        merged = {}
        if components:
            for key in set(components).union(mdapi_components):
                merged[key] = list(
                    set(chain(components.get(key, []), mdapi_components.get(key, [])))
                )
            components = merged
        else:
            components = mdapi_components

        if api_retrieve_unpackaged_response:
            api_retrieve_unpackaged_response.extend(mdapi_response_messages)
        else:
            api_retrieve_unpackaged_response = mdapi_response_messages

        return [components, api_retrieve_unpackaged_response]

    def _copy_to_tempdir(self, src_dir, temp_dir):
        for item in os.listdir(src_dir):
            src_item = os.path.join(src_dir, item)
            dst_item = os.path.join(temp_dir, item)

            if os.path.isdir(src_item):
                if not os.path.exists(dst_item):
                    shutil.copytree(src_item, dst_item)
                else:
                    self._merge_directories(src_item, dst_item)
            else:
                if not os.path.exists(dst_item):
                    shutil.copy2(src_item, dst_item)
                else:
                    self.logger.debug(f"File {dst_item} already exists, skipping...")

    def _merge_directories(self, src_dir, dst_dir):
        for item in os.listdir(src_dir):
            src_item = os.path.join(src_dir, item)
            dst_item = os.path.join(dst_dir, item)

            if os.path.isdir(src_item):
                if not os.path.exists(dst_item):
                    shutil.copytree(src_item, dst_item)
                    self._merge_directories(src_item, dst_item)
            else:
                if not os.path.exists(dst_item):
                    shutil.copy2(src_item, dst_item)  # Copy file if it doesn't exist
                else:
                    self.logger.debug(f"File {dst_item} already exists, skipping...")

    def _is_plan(self, name):

        if self.project_config.lookup(f"plans__{name}") is not None:
            return True
        elif self.project_config.lookup(f"flows__{name}") is not None:
            return False
        else:
            raise TaskOptionsError(
                "No paths provided and unable to determine the current plan or flow name."
            )

    def _get_plan_tasks(self, name, is_plan=False):

        tasks = []
        if is_plan:
            step_config = self.project_config.lookup(f"plans__{name}")
        else:
            step_config = self.project_config.lookup(f"flows__{name}")

        tasks = self._freeze_steps(self.project_config, step_config)

        return tasks

    def _freeze_steps(self, project_config, plan_config) -> list:
        steps = plan_config["steps"]
        flow_config = FlowConfig(plan_config)
        flow_config.project_config = project_config
        flow = FlowCoordinator(project_config, flow_config)
        steps = []
        for step in flow.steps:
            if step.skip:
                continue
            with cd(step.project_config.repo_root):
                task = step.task_class(
                    step.project_config,
                    TaskConfig(step.task_config),
                    name=step.task_name,
                )
                steps.extend(task.freeze(step))
        self.logger.debug("Prepared steps:\n" + json.dumps(steps, indent=4))
        return steps

    def _load_deploy_paths(self, name, is_plan=False):
        tasks = self._get_plan_tasks(name, is_plan)
        if tasks:
            self.deploy_paths = self._get_deployable_paths(tasks)

    def _get_deployable_paths(self, tasks):
        found_paths = []
        paths_to_search = ["path", "subfolder"]
        for task in tasks:
            if "task_config" in task and "options" in task["task_config"]:
                options = task["task_config"]["options"]
                found_paths.extend(self._search_for_paths(options, paths_to_search))
        return found_paths

    def _search_for_paths(self, options, keys_to_search):
        found_values = []

        if not keys_to_search:
            return found_values

        def recursive_search(obj):
            if isinstance(obj, dict):
                for key, value in obj.items():
                    if key in keys_to_search and isinstance(value, str):
                        found_values.append(value)
                    elif isinstance(value, (dict, list)):
                        recursive_search(value)

            elif isinstance(obj, list):
                for item in obj:
                    recursive_search(item)

        recursive_search(options)

        return found_values

    def _collect_components_from_paths(self, full_path):

        if not os.path.exists(full_path):
            return None, None

        components = defaultdict(set)
        self.logger.info(f"Collecting components from path: {full_path}")
        # remove if any exiting package.xml files coppied from deploy_pre/post paths

        if os.path.exists(os.path.join(full_path, "package.xml")):
            os.remove(os.path.join(full_path, "package.xml"))

        with convert_sfdx_source(full_path, None, self.logger) as src_path:
            package_xml_path = os.path.join(src_path, "package.xml")
            if os.path.exists(package_xml_path):
                try:
                    source_xml_tree = metadata_tree.parse(package_xml_path)
                    self.logger.info("parsing package.xml")

                    components = metadata_tree.parse_package_xml_types(
                        "name", source_xml_tree
                    )

                    response_messages = self._get_api_object_responce(
                        package_xml_path, source_xml_tree.version.text
                    )

                    return [components, response_messages]

                except ParseError as e:
                    self.logger.error(f"Error parsing package.xml: {e}")
                    return None, None
            else:
                self.logger.warning(
                    f"No package.xml found in {full_path}, scanning directories"
                )

        return None, None

    def _get_api_object_responce(self, pakcage_xml_path, version):

        if not os.path.exists(pakcage_xml_path):
            return None

        package_xml = open(pakcage_xml_path, "r")

        api_retrieve_unpackaged_object = self.api_retrieve_unpackaged(
            self, package_xml.read(), version
        )

        response_messages = parseString(
            api_retrieve_unpackaged_object._get_response().content
        ).getElementsByTagName("messages")

        return response_messages



# For backwards-compatibility
import cumulusci.tasks.vcs.pull_request as PullRequestsTask
from cumulusci.utils.deprecation import warn_moved


class PullRequests(PullRequestsTask.PullRequests):
    """Deprecated: use cumulusci.tasks.vcs.pull_request.PullRequests instead"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        warn_moved("cumulusci.tasks.vcs.pull_request", __name__)



from unittest import mock

import pytest

from cumulusci.core.config import OrgConfig


@pytest.fixture
def devhub_config():
    org_config = OrgConfig(
        {"instance_url": "https://devhub.my.salesforce.com", "access_token": "token"},
        "devhub",
    )
    org_config.refresh_oauth_token = mock.Mock()

    return org_config


@pytest.fixture
def org_config():
    org_config = OrgConfig(
        {
            "instance_url": "https://scratch.my.salesforce.com",
            "access_token": "token",
            "config_file": "orgs/scratch_def.json",
        },
        "dev",
    )
    org_config.refresh_oauth_token = mock.Mock()
    return org_config



import copy
import gc
import json
import os
import random
import sys
import tracemalloc
from contextlib import contextmanager, nullcontext
from functools import lru_cache
from pathlib import Path
from typing import Optional
from unittest import mock

import responses
import yaml
from requests import ReadTimeout

from cumulusci.core.config import (
    BaseConfig,
    BaseProjectConfig,
    OrgConfig,
    UniversalConfig,
)
from cumulusci.core.keychain import BaseProjectKeychain

# putting this below FakeBulkAPI causes a circular import
CURRENT_SF_API_VERSION = UniversalConfig().project__package__api_version
from cumulusci.tasks.bulkdata.tests.utils import FakeBulkAPI


def random_sha():
    hash = random.getrandbits(128)
    return "%032x" % hash


def create_project_config(
    repo_name="TestRepo",
    repo_owner="TestOwner",
    repo_commit=None,
    namespace=None,
):
    universal_config = UniversalConfig()
    project_config = DummyProjectConfig(
        universal_config=universal_config,
        repo_name=repo_name,
        repo_owner=repo_owner,
        repo_commit=repo_commit,
        config=copy.deepcopy(universal_config.config),
    )
    if namespace:
        project_config.config["project"]["package"]["namespace"] = namespace

    if project_config.config["services"] is None:
        project_config.config["services"] = {}

    project_config.config["services"]["test_service"] = {
        "attributes": {"name": {"required": True}, "password": {}},
    }
    keychain = BaseProjectKeychain(project_config, None)
    project_config.set_keychain(keychain)
    return project_config


class DummyProjectConfig(BaseProjectConfig):
    def __init__(
        self, universal_config, repo_name, repo_owner, repo_commit=None, config=None
    ):
        repo_info = {
            "owner": repo_owner,
            "name": repo_name,
            "url": f"https://github.com/{repo_owner}/{repo_name}",
            "commit": repo_commit or random_sha(),
        }
        super(DummyProjectConfig, self).__init__(
            universal_config, config, repo_info=repo_info
        )


DEFAULT_CONFIG = {
    "instance_url": "https://orgname.my.salesforce.com",
    "access_token": "pytest_sf_orgconnect_abc123",
    "id": "https://test.salesforce.com/id/00D0xORGID00000000/USERID",
    "username": "sfuser@example.com",
}


class DummyOrgConfig(OrgConfig):
    def __init__(self, config=None, name=None, keychain=None, global_org=False):
        config = {**DEFAULT_CONFIG, **(config or {})}
        name = name or "test"
        self._installed_packages = {}
        super(DummyOrgConfig, self).__init__(config, name, keychain, global_org)

    def refresh_oauth_token(self, keychain):
        pass


class DummyLogger(object):
    def __init__(self):
        self.out = []

    def log(self, msg, *args):
        self.out.append(msg % args)

    # Compatibility with various logging methods like info, warning, etc
    def __getattr__(self, name):
        return self.log

    def get_output(self):
        return "\n".join(self.out)


class DummyService(BaseConfig):
    password = "dummy_password"
    client_id = "ZOOMZOOM"

    def __init__(self, name, alias):
        self.name = name
        super().__init__()


class DummyKeychain(BaseProjectKeychain):
    def __init__(self, global_config_dir=None, cache_dir=None, project_config=None):
        self._global_config_dir = global_config_dir
        self._cache_dir = cache_dir
        if not project_config:
            project_config = create_project_config()
            project_config.keychain = self
        super().__init__(project_config, "XYZZY")

    @property
    def global_config_dir(self):
        return self._global_config_dir or Path.home() / "cumulusci"

    @property
    def cache_dir(self):
        return self._cache_dir or Path.home() / "project/.cci"

    def get_service(self, name, alias=None):
        return DummyService(name, alias)

    def set_org(self, org: OrgConfig, global_org: bool):
        pass


@contextmanager
def assert_max_memory_usage(max_usage):
    "Assert that a test does not exceed a certain memory threshold"
    tracemalloc.start()
    yield
    current, peak = tracemalloc.get_traced_memory()
    if peak > max_usage:
        big_objs(traced_only=True)
        assert peak < max_usage, ("Peak incremental memory usage was high:", peak)
    tracemalloc.stop()


def big_objs(traced_only=False):

    big_objs = (
        (sys.getsizeof(obj), obj)
        for obj in gc.get_objects()
        if sys.getsizeof(obj) > 20000
        and (tracemalloc.get_object_traceback(obj) if traced_only else True)
    )
    for size, obj in big_objs:
        print(type(obj), size, tracemalloc.get_object_traceback(obj))


class FakeSObjectProxy:
    def __init__(self, describe_data):
        self.describe_data = describe_data

    def describe(self):
        return self.describe_data


class FakeSF:
    """Simplistic mock of the Simple-Salesforce API

    Can be improved as needed over time.
    """

    fakes = {}
    headers = {}
    session = mock.Mock()
    base_url = "https://fakesf.example.org/"

    def describe(self):
        return self._get_json("Global")

    @property
    def sf_version(self):
        return CURRENT_SF_API_VERSION

    def _get_json(self, fake_dataset):
        self.fakes[fake_dataset] = self.fakes.get(fake_dataset, None) or json.loads(
            read_mock(fake_dataset)
        )
        return self.fakes[fake_dataset]

    def __getattr__(self, name):
        return FakeSObjectProxy(self._get_json(name))


@lru_cache  # change to @cache when Python 3.9 is allowed
def read_mock(name: str):
    base_path = Path(__file__).parent.parent / "tests/shared_cassettes"

    with (base_path / f"GET_sobjects_{name}_describe.yaml").open("r") as f:
        return yaml.safe_load(f)["response"]["body"]["string"]


def mock_describe_calls(domain="example.com", version=CURRENT_SF_API_VERSION):
    def mock_sobject_describe(name: str):
        responses.add(
            method="GET",
            url=f"https://{domain}/services/data/v{version}/sobjects/{name}/describe",
            body=read_mock(name),
            status=200,
        )

    responses.add(
        method="GET",
        url=f"https://{domain}/services/data",
        body=json.dumps([{"version": version}]),
        status=200,
    )
    responses.add(
        method="GET",
        url=f"https://{domain}/services/data",
        body=json.dumps([{"version": version}]),
        status=200,
    )

    responses.add(
        method="GET",
        url=f"https://{domain}/services/data/v{version}/sobjects",
        body=read_mock("Global"),
        status=200,
    )

    for sobject in [
        "Account",
        "Contact",
        "Opportunity",
        "OpportunityContactRole",
        "Case",
        "Event",
    ]:
        mock_sobject_describe(sobject)


@contextmanager
def mock_salesforce_client(task, *, is_person_accounts_enabled=False):
    mock_describe_calls("test.salesforce.com")

    real_init = task._init_task
    salesforce_client = FakeSF()

    def _init_task():
        real_init()
        task.bulk = FakeBulkAPI()
        task.sf = salesforce_client

    with mock.patch(
        "cumulusci.core.config.org_config.OrgConfig.is_person_accounts_enabled",
        lambda: is_person_accounts_enabled,
    ), mock.patch.object(task, "_init_task", _init_task):
        yield


# copy over most of the environment because Windows needs something
# non-obvious and it is quite laborius to figure out what it is.

# but we want to hide the CUMULUSCI_ and GITHUB_ related stuff from tests.
ENV_CLONE = {
    key: value
    for key, value in os.environ.items()
    if "CUMULUSCI_" not in key and "GITHUB_" not in key
}

ENV_CLONE["CUMULUSCI_SYSTEM_CERTS"] = "True"


@contextmanager
def mock_env(
    home,
    cumulusci_key: Optional[str] = None,
):
    cumulusci_key = cumulusci_key if cumulusci_key else "0123456789ABCDEF"
    real_homedir = str(Path.home())
    new_environment = {
        "HOME": home,
        "USERPROFILE": home,
        "REAL_HOME": real_homedir,
        "CUMULUSCI_KEY": cumulusci_key,
        **ENV_CLONE,
    }

    with mock.patch("pathlib.Path.home", lambda: Path(home)), mock.patch.dict(
        os.environ, new_environment, clear=True
    ):
        yield


def unmock_env():
    """Reset homedir and CCI environment variable or leave them if they weren't changed"""
    if "REAL_HOME" in os.environ:
        cci_key = os.environ.get("REAL_CUMULUSCI_KEY") or None
        homedir = os.environ["REAL_HOME"]
        return mock_env(homedir, cci_key)
    else:
        return nullcontext()


class FakeUnreliableRequestHandler:
    """Fake a request handler which fails its second request."""

    counter = 0

    def __init__(self, response=None, exception=ReadTimeout):
        self.response = response
        self.exception = exception

    def request_callback(self, request):
        should_return_error = self.counter == 1  # fail the second request of X
        self.counter += 1
        if should_return_error:
            raise self.exception()
        else:
            return (
                200,
                {"Last-Modified": "Wed, 01 Jan 2000 01:01:01 GMT"},
                json.dumps(self.real_reliable_request_callback(request)),
            )

    def real_reliable_request_callback(self, request):
        return self.response


CURRENT_SF_API_VERSION = (
    CURRENT_SF_API_VERSION  # quiet linter and export to other modules
)



import os
import time
from unittest import mock

import pytest

from cumulusci.core.config import BaseProjectConfig, TaskConfig, UniversalConfig
from cumulusci.core.exceptions import TaskOptionsError
from cumulusci.tasks.metadata.ee_src import CreateUnmanagedEESrc, RevertUnmanagedEESrc
from cumulusci.utils import temporary_dir


class TestCreateUnmanagedEESrc:
    @mock.patch("cumulusci.tasks.metadata.ee_src.remove_xml_element_directory")
    def test_run_task(self, remove_xml_element_directory):
        with temporary_dir() as path:
            revert_path = os.path.join(
                os.path.dirname(path), os.path.basename(path) + "_revert"
            )
            project_config = BaseProjectConfig(
                UniversalConfig(), config={"noyaml": True}
            )
            task_config = TaskConfig(
                {"options": {"path": path, "revert_path": revert_path}}
            )
            task = CreateUnmanagedEESrc(project_config, task_config)
            task()
            remove_xml_element_directory.assert_has_calls(
                [
                    mock.call("availableFields", path, "*.object"),
                    mock.call("visibility[.='Protected']", path, "*.object"),
                ]
            )

    def test_run_task__path_not_found(self):
        project_config = BaseProjectConfig(UniversalConfig(), config={"noyaml": True})
        task_config = TaskConfig({"options": {"path": "bogus", "revert_path": "bogus"}})
        task = CreateUnmanagedEESrc(project_config, task_config)
        with pytest.raises(TaskOptionsError):
            task()

    def test_run_task__revert_path_already_exists(self):
        with temporary_dir() as path, temporary_dir() as revert_path:
            project_config = BaseProjectConfig(
                UniversalConfig(), config={"noyaml": True}
            )
            task_config = TaskConfig(
                {"options": {"path": path, "revert_path": revert_path}}
            )
            task = CreateUnmanagedEESrc(project_config, task_config)
            with pytest.raises(TaskOptionsError):
                task()


class TestRevertUnmanagedEESrc:
    def test_run_task(self, tmp_path):
        revert_path = tmp_path / "revert"
        revert_path.mkdir()
        file_path = revert_path / "file"
        file_path.write_text("content")

        path = tmp_path / "path"
        path.mkdir()
        project_config = BaseProjectConfig(UniversalConfig(), config={"noyaml": True})
        task_config = TaskConfig(
            {"options": {"path": str(path), "revert_path": str(revert_path)}}
        )
        task = RevertUnmanagedEESrc(project_config, task_config)
        task()
        assert (path / "file").exists()

    def test_run_task__revert_path_not_found(self):
        project_config = BaseProjectConfig(UniversalConfig(), config={"noyaml": True})
        task_config = TaskConfig({"options": {"path": "bogus", "revert_path": "bogus"}})
        task = RevertUnmanagedEESrc(project_config, task_config)
        with pytest.raises(TaskOptionsError):
            task()

    def test_revert_with_update(self, tmp_path):
        """
        Test the 'update' behavior of RevertUnmanagedEESrc task with temporary directories.

        This test creates a source and a destination directory each with one
        file. The file in the source directory has an older timestamp. After
        running RevertUnmanagedEESrc, it checks that the destination file is not
        overwritten by the older source file, confirming the update logic.
        """
        source_dir = tmp_path / "source"
        source_dir.mkdir()
        source_file = source_dir / "testfile.txt"
        source_file.write_text("original content")

        dest_dir = tmp_path / "dest"
        dest_dir.mkdir()
        dest_file = dest_dir / "testfile.txt"
        dest_file.write_text("modified content")

        # Ensure the source file has an older timestamp
        past_time = time.time() - 100
        # Use os.utime to modify the timestamp
        source_file.touch()
        os.utime(str(source_file), (past_time, past_time))

        project_config = BaseProjectConfig(UniversalConfig(), config={"noyaml": True})
        task_config = TaskConfig(
            {"options": {"path": str(dest_dir), "revert_path": str(source_dir)}}
        )
        task = RevertUnmanagedEESrc(project_config, task_config)
        task()

        # Verify that the destination file was not updated (due to older source file)
        assert dest_file.read_text() == "modified content"



import os
import sys
import warnings
from importlib.metadata import PackageNotFoundError, version

# Suppress pkg_resources deprecation warnings from dependencies
warnings.filterwarnings("ignore", message=".*pkg_resources.*", category=UserWarning)

from simple_salesforce import api, bulk

__location__ = os.path.dirname(os.path.realpath(__file__))

try:
    __version__ = version("cumulusci-plus")
except PackageNotFoundError:
    __version__ = "unknown"

try:
    version("cumulusci")
    raise Exception("CumulusCI installation found, Remove the CumulusCI package.")
except PackageNotFoundError:
    pass

if sys.version_info < (3, 11):  # pragma: no cover
    raise Exception("CumulusCI requires Python 3.11+.")

api.OrderedDict = dict
bulk.OrderedDict = dict



import os
import shutil
from abc import ABC, abstractmethod
from typing import Type

import fs

from cumulusci.core.exceptions import DependencyResolutionError, VcsNotFoundError
from cumulusci.core.utils import import_global
from cumulusci.utils import download_extract_vcs_from_repo
from cumulusci.utils.yaml.cumulusci_yml import VCSSourceModel, VCSSourceRelease

# To avoid circular dependency error
# from cumulusci.core.config import BaseProjectConfig
# from cumulusci.vcs.base import VCSService
# from cumulusci.vcs.models import AbstractRepo


class VCSSource(ABC):
    """Abstract base class for VCS sources.
    This class defines the interface for VCS sources and provides
    common functionality for working with VCS sources.
    Subclasses should implement the methods and properties defined here.
    """

    _registry = {}
    # project_config : BaseProjectConfig
    spec: VCSSourceModel
    vcs: str
    url: str
    location: str
    # repo: AbstractRepo
    repo_owner: str
    repo_name: str
    # vcs_service: VCSService

    def __init__(self, project_config, spec: VCSSourceModel):
        self.project_config = project_config
        self.spec = spec
        self.vcs = spec.vcs
        self.url = spec.url
        self.location = self.url

        if self.url.endswith(".git"):
            self.url = self.url[:-4]

        self._resolve_repo()

    def _resolve_repo(self):
        from cumulusci.vcs.base import VCSService

        self.vcs_service: VCSService = self.get_vcs_service()

        try:
            from cumulusci.vcs.models import AbstractRepo

            self.repo: AbstractRepo = self.vcs_service.get_repository(
                options={"repository_url": self.url}
            )
        except VcsNotFoundError:
            raise DependencyResolutionError(
                f"We are unable to find the repository at {self.url}. Please make sure the URL is correct, that your GitHub user has read access to the repository, and that your GitHub personal access token includes the “repo” scope."
            )

        self._set_additional_repo_config()

        self.resolve()

    @classmethod
    def register(cls, vcs_type: str, subclass: str):
        """Register a subclass of VCSSource with a specific type name.
        This allows the VCSSource class to create instances of the subclass
        based on the type name specified in the VCSSourceModel.
        Args:
            vcs_type (str): The vcs type name to register the subclass with, ex: 'github'.
            subclass (str): The subclass to register, ex: 'cumulusci.core.source.GitHubSource'.
        """
        cls._registry[vcs_type] = subclass

    @classmethod
    def registered_source_models(cls):
        """Returns a list of registered classes.
        return:
            [GitHubSource, ...].
        """
        return_list = []
        for source_klass_path in cls._registry.values():
            if source_klass_path:
                source_klass = import_global(source_klass_path)
                if issubclass(source_klass, VCSSource):
                    return_list.append(source_klass.source_model())
        return return_list

    @classmethod
    def create(cls, project_config, spec: VCSSourceModel):
        source_klass_path = cls._registry.get(spec.vcs, None)
        if source_klass_path:
            source_klass = import_global(source_klass_path)
            if issubclass(source_klass, VCSSource):
                return source_klass(project_config, spec)

        raise ValueError(f"No child class found for type_name={spec.vcs}")

    @classmethod
    @abstractmethod
    def source_model(self) -> Type[VCSSourceModel]:
        raise NotImplementedError("Subclasses should implement source_model")

    @abstractmethod
    def get_vcs_service(self):
        raise NotImplementedError("Subclasses should implement get_vcs_service")

    @abstractmethod
    def __repr__(self):
        raise NotImplementedError(
            "Subclasses should implement __repr__ to provide a string representation of the VCS source."
        )

    @abstractmethod
    def __str__(self):
        raise NotImplementedError(
            "Subclasses should implement __str__ to provide a string representation of the VCS source."
        )

    @abstractmethod
    def __hash__(self):
        raise NotImplementedError(
            "Subclasses should implement __hash__ to provide a hash value for the VCS source."
        )

    @abstractmethod
    def get_ref(self):
        raise NotImplementedError(
            "Subclasses should implement get_ref to retrieve the reference information."
        )

    @abstractmethod
    def get_tag(self):
        raise NotImplementedError(
            "Subclasses should implement get_tag to retrieve the tag information."
        )

    @abstractmethod
    def get_branch(self):
        raise NotImplementedError(
            "Subclasses should implement get_branch to retrieve the branch information."
        )

    @abstractmethod
    def get_release_tag(self):
        raise NotImplementedError(
            "Subclasses should implement get_release_tag to retrieve the release tag information."
        )

    def _set_additional_repo_config(self):
        ### Subclasses can override this method to set additional configuration on the repo.
        """Set additional configuration on the repo."""
        pass

    def resolve(self):
        from cumulusci.vcs.bootstrap import find_latest_release, find_previous_release

        """Resolve a VCS source into a specific commit.

        The spec must include:
        - vcs: the VCS repository type (e.g., GitHub, Bitbucket, ADO)
        - url: the URL of the VCS repository

        It's recommended that the source include:
        - resolution_strategy: [production | preproduction | <strategy-name>]

        The spec may instead be specific about the desired ref or release:
        - commit: a commit hash
        - ref: a Git ref
        - branch: a Git branch
        - tag: a Git tag
        - release: "latest" | "previous" | "latest_beta"

        If none of these are specified, CumulusCI will use the resolution strategy "production"
        to locate the appropriate release or ref.
        """
        ref = None
        self.branch = None
        # These branches preserve some existing behavior: when a source is set to
        # a specific tag or release, there's no fallback, as there would be
        # if we were subsumed within the dependency resolution machinery.

        # If the user was _not_ specific, we will use the full resolution stack,
        # including fallback behaviors.
        if self.spec.commit:
            self.commit = self.description = self.spec.commit
            return
        elif self.spec.ref:
            ref = self.get_ref()
        elif self.spec.tag:
            ref = self.get_tag()
        elif self.spec.branch:
            self.branch = self.spec.branch
            ref = self.get_branch()
        elif self.spec.release:
            release = None
            if self.spec.release is VCSSourceRelease.LATEST:
                release = find_latest_release(self.repo, include_beta=False)
            elif self.spec.release is VCSSourceRelease.LATEST_BETA:
                release = find_latest_release(self.repo, include_beta=True)
            elif self.spec.release is VCSSourceRelease.PREVIOUS:
                release = find_previous_release(self.repo)
            if release is None:
                raise DependencyResolutionError(
                    f"Could not find release {self.spec.release}."
                )
            ref = release.tag_ref_name
        else:
            # Avoid circular import issues
            from cumulusci.core.dependencies.resolvers import (
                get_resolver_stack,
                resolve_dependency,
            )

            dynamic_dependency_cls = self.vcs_service.dynamic_dependency_class

            # Use resolution strategies to find the right commit.
            dep = dynamic_dependency_cls(url=self.spec.url)
            dep.set_repo(self.repo)

            resolve_dependency(
                dep,
                self.project_config,
                get_resolver_stack(self.project_config, self.spec.resolution_strategy),
            )

            self.commit = self.description = dep.ref
            return

        self.description = ref[6:] if ref.startswith("heads/") else ref
        self.commit = self.repo.get_ref(ref).sha

    def fetch(self):
        """Fetch the archive of the specified commit and construct its project config."""
        with self.project_config.open_cache(
            fs.path.join("projects", self.repo.repo_name, self.commit)
        ) as path:
            zf = download_extract_vcs_from_repo(self.repo, ref=self.commit)
            try:
                zf.extractall(path)
            except Exception:
                # make sure we don't leave an incomplete cache
                shutil.rmtree(path)
                raise

        project_config = self.project_config.construct_subproject_config(
            repo_info={
                "root": os.path.realpath(path),
                "owner": self.repo.repo_owner,
                "name": self.repo.repo_name,
                "url": self.url,
                "commit": self.commit,
                # Note: we currently only pass the branch if it was explicitly
                # included in the source spec. If the commit was found another way,
                # we aren't looking up what branches have that commit as their head.
                "branch": self.branch,
            }
        )
        return project_config

    @property
    def frozenspec(self):
        """Return a spec to reconstruct this source at the current commit"""
        # TODO: The branch name is lost when freezing the source for MetaDeploy.
        # We could include it here, but it would fail validation when GitHubSourceModel
        # parses it due to having both commit and branch.
        return {
            "vcs": self.vcs,
            "url": self.url,
            "commit": self.commit,
            "description": self.description,
        }

    @property
    def allow_remote_code(self) -> bool:
        return self.spec.allow_remote_code


VCSSource.register("github", "cumulusci.core.source.github.GitHubSource")
VCSSource.register(
    "github_enterprise", "cumulusci.core.source.github.GitHubEnterpriseSource"
)



import typing as T

from cumulusci.core.exceptions import ServiceNotConfigured


class SubprocessKeychain(T.NamedTuple):
    """A pretend, in-memory keychain that knows about connected apps and nothing else.

    This keychain is used by Snowfakery to make the connected app available to
    tasks that are running in sub-processes. If Snowfakery ever needs a broader
    range of services thaan just the connected app, this keychain will need to
    become more complex.
    """

    connected_app: T.Any = None

    def get_service(self, name, alias=None):
        if name == "connected_app" and self.connected_app:
            return self.connected_app

        raise ServiceNotConfigured(name)

    def set_org(self, *args):
        pass



import logging
from abc import ABC, abstractmethod
from typing import List, Optional, Type

from cumulusci.core.config import BaseProjectConfig, ServiceConfig
from cumulusci.core.dependencies.base import DynamicDependency
from cumulusci.core.keychain import BaseProjectKeychain
from cumulusci.tasks.release_notes.generator import BaseReleaseNotesGenerator
from cumulusci.utils.classutils import get_all_subclasses
from cumulusci.vcs.models import AbstractRelease, AbstractRepo
from cumulusci.vcs.utils import AbstractCommitDir


class VCSService(ABC):
    """This is an abstract base class for VCS services.
    It defines the interface that all VCS services must implement.
    Subclasses should provide their own implementations of the methods and properties defined here.
    """

    logger: logging.Logger
    config: BaseProjectConfig
    service_config: ServiceConfig
    name: str
    keychain: Optional[BaseProjectKeychain]
    _service_registry: List["VCSService"] = []

    def __init__(
        self, config: BaseProjectConfig, name: Optional[str] = None, **kwargs
    ) -> None:
        """Initializes the VCS service with the given configuration, service name, and keychain.

        Args:
            config (BaseProjectConfig): The configuration for the GitHub service.
            name (str): Optional: The name or alias of the VCS service.
            **kwargs: Additional keyword arguments.
        """
        self.config = config
        self.service_config = config.keychain.get_service(self.service_type, name)
        self.name = self.service_config.name or name
        self.keychain = config.keychain
        self.logger = kwargs.get("logger") or logging.getLogger(__name__)

    @property
    def service_type(self) -> str:
        """Returns the service type of the VCS service.
        This property should be overridden by subclasses to provide
        the specific service type. For example, it could return "github",
        "bitbucket", etc. The service type is used to identify the
        specific VCS service being used."""
        if isinstance(self.__class__.service_type, property):
            raise NotImplementedError(
                "Subclasses should define the service_type property"
            )
        return self.__class__.service_type

    @property
    @abstractmethod
    def dynamic_dependency_class(self) -> Type[DynamicDependency]:
        """Returns the dynamic dependency class for the VCS service.
        This property should be overridden by subclasses to provide
        the specific dynamic dependency class. For example, it could
        return "GitHubDynamicDependency", "BitbucketDynamicDependency", etc."""
        raise NotImplementedError(
            "Subclasses should define the dynamic_dependency_class property"
        )

    @classmethod
    @abstractmethod
    def validate_service(cls, options: dict, keychain) -> dict:
        """Validate the service configuration.
        This method should be overridden by subclasses to provide
        specific validation logic. For example, it could check if the
        required options are present, valid values and able to
        establish a connection with the VCS.
        The method should raise an exception if the validation fails."""
        raise NotImplementedError("Subclasses should provide their own implementation")

    @classmethod
    @abstractmethod
    def get_service_for_url(
        cls, project_config: BaseProjectConfig, url: str, service_alias: str = None
    ) -> Optional["VCSService"]:
        """Returns the service configuration for the given URL.
        This method should be overridden by subclasses to provide
        specific logic for retrieving the service configuration.
        The method should raise an exception if the validation fails."""
        raise NotImplementedError("Subclasses should provide their own implementation")

    @classmethod
    def registered_services(cls) -> List[Type["VCSService"]]:
        """This method returns all subclasses of VCSService that have been registered.
        It can be used to dynamically discover available VCS services."""
        return get_all_subclasses(cls)

    @abstractmethod
    def get_repository(self, options: dict = {}) -> AbstractRepo:
        """Returns the repository object for the VCS service.
        This method should be overridden by subclasses to provide
        the specific implementation for retrieving the repository.
        The method should return an instance of a class that implements
        the AbstractRepo interface."""
        raise NotImplementedError("Subclasses should provide their own implementation")

    @abstractmethod
    def parse_repo_url(self) -> List[str]:
        """Returns the owner, repo_name, and host from the repository URL.
        This method should be overridden by subclasses to provide
        the specific implementation for parsing the repository URL.
        The method should return a list containing the owner, repo_name, and host."""
        raise NotImplementedError(
            "Subclasses should provide their own implementation of parse_repo_url"
        )

    @abstractmethod
    def get_committer(self, repo: AbstractRepo) -> AbstractCommitDir:
        """Returns the committer for the VCS repository."""
        raise NotImplementedError("Subclasses should provide their own implementation")

    @abstractmethod
    def markdown(
        self, release: AbstractRelease, mode: str = "", context: str = ""
    ) -> str:
        """Returns the markdown for the release.
        This method should be overridden by subclasses to provide
        the specific implementation for retrieving the markdown.
        The method should return a string containing the markdown."""
        raise NotImplementedError("Subclasses should provide their own implementation")

    @abstractmethod
    def release_notes_generator(self, options: dict) -> BaseReleaseNotesGenerator:
        """Returns the release notes generator for the VCS service."""
        raise NotImplementedError(
            "Subclasses should define the release_notes_generator property"
        )

    @abstractmethod
    def parent_pr_notes_generator(
        self, repo: AbstractRepo
    ) -> BaseReleaseNotesGenerator:
        """Returns the parent PR notes generator for the VCS service."""
        raise NotImplementedError(
            "Subclasses should define the parent_pr_notes_generator property"
        )



import importlib.metadata
import logging
from functools import lru_cache
from typing import Dict, List, Optional, Type

from cumulusci.plugins.plugin_base import PluginBase

logger: logging.Logger = logging.getLogger(__name__)


def get_plugin_manager() -> "PluginManager":
    """Get the plugin manager instance."""
    return PluginManager()


@lru_cache(maxsize=50)
def load_plugins() -> List[PluginBase]:
    """Load all available plugins and return them as a list."""
    manager = get_plugin_manager()
    plugins = []
    for name, plugin_class in manager._plugins.items():
        try:
            plugin = plugin_class()
            plugin.initialize()
            plugins.append(plugin)
        except Exception as e:
            logger.warning(f"Failed to initialize plugin {name}: {str(e)}")
    return plugins


class PluginManager:
    """Manages the loading and access of CumulusCI plugins."""

    def __init__(self) -> None:
        self._plugins: Dict[str, Type[PluginBase]] = {}
        self.logger = logging.getLogger(__name__)
        self._load_plugins()

    def _load_plugins(self) -> None:
        """Load all available plugins."""
        try:
            for entry_point in importlib.metadata.entry_points().select(
                group="cumulusci.plugins"
            ):
                try:
                    plugin_class = entry_point.load()
                    self._plugins[entry_point.name] = plugin_class
                except Exception as e:
                    self.logger.warning(
                        f"Failed to load plugin {entry_point.name}: {str(e)}"
                    )
        except Exception as e:
            self.logger.warning(f"Failed to load plugins: {str(e)}")

    def get_plugin(self, name: str) -> Optional[Type[PluginBase]]:
        """Get a plugin by name."""
        return self._plugins.get(name)

    def list_plugins(self) -> List[str]:
        """List all available plugin names."""
        return list(self._plugins.keys())



from cumulusci.core.exceptions import TaskOptionsError
from cumulusci.core.utils import process_list_arg
from cumulusci.tasks.metadata_etl import MetadataSingleEntityTransformTask
from cumulusci.utils.xml.metadata_tree import MetadataElement

OBJ_SETTINGS = [
    "Activities",
    "BulkApi",
    "Feeds",
    "History",
    "Licensing",
    "Reports",
    "Search",
    "Sharing",
    "StreamingApi",
]
OBJ_SETTINGS_STR = ", ".join(OBJ_SETTINGS)


class SetObjectSettings(MetadataSingleEntityTransformTask):
    entity = "CustomObject"
    task_options = {
        "enable": {
            "description": f"Array of object settings to enable. Uses the setting name.  Available values: {OBJ_SETTINGS_STR}"
        },
        "disable": {
            "description": f"Array of object settings to disable. Uses the setting name.  Available values: {OBJ_SETTINGS_STR}"
        },
        **MetadataSingleEntityTransformTask.task_options,
    }

    def _init_options(self, kwargs):
        super()._init_options(kwargs)
        if not self.options.get("enable") and not self.options.get("disable"):
            raise TaskOptionsError(
                "You must provide values for either 'enable' or 'disable'"
            )
        self.options["enable"] = self._process_settings(self.options.get("enable"))
        self.options["disable"] = self._process_settings(self.options.get("disable"))

    def _process_settings(self, settings):
        settings = process_list_arg(settings)
        if settings is None:
            settings = []
        invalid = []
        for setting in settings:
            if setting not in OBJ_SETTINGS:
                invalid.append(setting)
        if invalid:
            invalid_settings = ", ".join(invalid)
            raise TaskOptionsError(
                f"Invalid settings: {invalid_settings}.  Valid settings are: {OBJ_SETTINGS_STR}"
            )
        return settings

    def _transform_entity(self, metadata: MetadataElement, api_name: str):
        for setting in self.options["enable"]:
            self._apply_setting(metadata, setting, True)
        for setting in self.options["disable"]:
            self._apply_setting(metadata, setting, False)
        return metadata

    def _apply_setting(self, metadata: MetadataElement, setting: str, enable: bool):
        name = f"enable{setting}"
        elem = metadata.find(name)
        if elem is None:
            elem = metadata.append(name)
        elem.text = "true" if enable else "false"



from simple_salesforce.exceptions import SalesforceResourceNotFound

from cumulusci.core.template_utils import format_str
from cumulusci.robotframework.base_library import BaseLibrary

# https://developer.salesforce.com/docs/atlas.en-us.api_rest.meta/api_rest/resources_composite_sobjects_collections_create.htm
SF_COLLECTION_INSERTION_LIMIT = 200
STATUS_KEY = ("status",)


class SalesforceAPI(BaseLibrary):
    """Keywords for interacting with Salesforce API"""

    def __init__(self):
        super().__init__()
        self._session_records = []

    def delete_session_records(self):
        """Deletes records that were created while running this test case.

        (Only records specifically recorded using the Store Session Record
        keyword are deleted.)
        """
        self._session_records.reverse()
        self.builtin.log("Deleting {} records".format(len(self._session_records)))
        for record in self._session_records[:]:
            self.builtin.log("  Deleting {type} {id}".format(**record))
            try:
                self.salesforce_delete(record["type"], record["id"])
            except SalesforceResourceNotFound:
                self.builtin.log("    {type} {id} is already deleted".format(**record))
            except Exception as e:
                self.builtin.log(
                    "    {type} {id} could not be deleted:".format(**record),
                    level="WARN",
                )
                self.builtin.log("      {}".format(e), level="WARN")

    def get_latest_api_version(self):
        """Return the API version used by the current org"""
        return self.cumulusci.org.latest_api_version

    def get_record_type_id(self, obj_type, developer_name):
        """Returns the Record Type Id for a record type name"""
        soql = "SELECT Id FROM RecordType WHERE SObjectType='{}' and DeveloperName='{}'".format(
            obj_type, developer_name
        )
        res = self.cumulusci.sf.query_all(soql)
        return res["records"][0]["Id"]

    def salesforce_delete(self, obj_name, obj_id):
        """Deletes a Salesforce object by object name and Id.

        Example:

        The following example assumes that ``${contact id}`` has been
        previously set. The example deletes the Contact with that Id.

        | Salesforce Delete  Contact  ${contact id}
        """
        self.builtin.log("Deleting {} with Id {}".format(obj_name, obj_id))
        obj_class = getattr(self.cumulusci.sf, obj_name)
        obj_class.delete(obj_id)
        self.remove_session_record(obj_name, obj_id)

    def salesforce_get(self, obj_name, obj_id):
        """Gets a Salesforce object by Id and returns the result as a dict.

        Example:

        The following example assumes that ``${contact id}`` has been
        previously set. The example retrieves the Contact object with
        that Id and then logs the Name field.

        | &{contact}=  Salesforce Get  Contact  ${contact id}
        | log  Contact name:  ${contact['Name']}

        """
        self.builtin.log(f"Getting {obj_name} with Id {obj_id}")
        obj_class = getattr(self.cumulusci.sf, obj_name)
        return obj_class.get(obj_id)

    def salesforce_insert(self, obj_name, **kwargs):
        """Creates a new Salesforce object and returns the Id.

        The fields of the object may be defined with keyword arguments
        where the keyword name is the same as the field name.

        The object name and Id is passed to the *Store Session
        Record* keyword, and will be deleted when the keyword
        *Delete Session Records* is called.

        As a best practice, either *Delete Session Records* or
        *Delete Records and Close Browser* from Salesforce.robot
        should be called as a suite teardown.

        Example:

        The following example creates a new Contact with the
        first name of "Eleanor" and the last name of "Rigby".

        | ${contact id}=  Salesforce Insert  Contact
        | ...  FirstName=Eleanor
        | ...  LastName=Rigby

        """
        self.builtin.log("Inserting {} with values {}".format(obj_name, kwargs))
        obj_class = getattr(self.cumulusci.sf, obj_name)
        res = obj_class.create(kwargs)
        self.store_session_record(obj_name, res["id"])
        return res["id"]

    def _salesforce_generate_object(self, obj_name, **fields):
        obj = {"attributes": {"type": obj_name}}  # Object type to create
        obj.update(fields)
        return obj

    def generate_test_data(self, obj_name, number_to_create, **fields):
        """Generate bulk test data

        This returns an array of dictionaries with template-formatted
        arguments which can be passed to the *Salesforce Collection Insert*
        keyword.

        You can use ``{{number}}`` to represent the unique index of
        the row in the list of rows.  If the entire string consists of
        a number, Salesforce API will treat the value as a number.

        Example:

        The following example creates three new Contacts:

            | @{objects} =  Generate Test Data  Contact  3
            | ...  Name=User {{number}}
            | ...  Age={{number}}

        The example code will generate Contact objects with these fields:

            | [{'Name': 'User 0', 'Age': '0'},
            |  {'Name': 'User 1', 'Age': '1'},
            |  {'Name': 'User 2', 'Age': '2'}]

        Python Expression Syntax is allowed so computed templates like this are also allowed: ``{{1000 + number}}``

        Python operators can be used, but no functions or variables are provided, so mostly you just
        have access to mathematical and logical operators. The Python operators are described here:

        https://www.digitalocean.com/community/tutorials/how-to-do-math-in-python-3-with-operators

        Contact the CCI team if you have a use-case that
        could benefit from more expression language power.

        Templates can also be based on faker patterns like those described here:

        https://faker.readthedocs.io/en/master/providers.html

        Most examples can be pasted into templates verbatim:

            | @{objects}=  Generate Test Data  Contact  200
            | ...  Name={{fake.first_name}} {{fake.last_name}}
            | ...  MailingStreet={{fake.street_address}}
            | ...  MailingCity=New York
            | ...  MailingState=NY
            | ...  MailingPostalCode=12345
            | ...  Email={{fake.email(domain="salesforce.com")}}

        """
        objs = []

        for i in range(int(number_to_create)):
            formatted_fields = {
                name: format_str(value, {"number": i}) for name, value in fields.items()
            }
            newobj = self._salesforce_generate_object(obj_name, **formatted_fields)
            objs.append(newobj)

        return objs

    def remove_session_record(self, obj_type, obj_id):
        """Remove a record from the list of records that should be automatically removed."""
        try:
            self._session_records.remove({"type": obj_type, "id": obj_id})
        except ValueError:
            self.builtin.log(
                "Did not find record {} {} in the session records list".format(
                    obj_type, obj_id
                )
            )

    def salesforce_collection_insert(self, objects):
        """Inserts records that were created with *Generate Test Data*.

        _objects_ is a list of data, typically generated by the
        *Generate Test Data* keyword.

        A 200 record limit is enforced by the Salesforce APIs.

        The object name and Id is passed to the *Store Session
        Record* keyword, and will be deleted when the keyword *Delete
        Session Records* is called.

        As a best practice, either *Delete Session Records* or
        **Delete Records and Close Browser* from Salesforce.robot
        should be called as a suite teardown.

        Example:

        | @{objects}=  Generate Test Data  Contact  200
        | ...  FirstName=User {{number}}
        | ...  LastName={{fake.last_name}}
        | Salesforce Collection Insert  ${objects}

        """
        assert (
            not obj.get("id", None) for obj in objects
        ), "Insertable objects should not have IDs"
        assert len(objects) <= SF_COLLECTION_INSERTION_LIMIT, (
            "Cannot insert more than %s objects with this keyword"
            % SF_COLLECTION_INSERTION_LIMIT
        )

        records = self.cumulusci.sf.restful(
            "composite/sobjects",
            method="POST",
            json={"allOrNone": True, "records": objects},
        )

        for idx, (record, obj) in enumerate(zip(records, objects)):
            if record["errors"]:
                raise AssertionError(
                    "Error on Object {idx}: {record} : {obj}".format(**vars())
                )
            self.store_session_record(obj["attributes"]["type"], record["id"])
            obj["id"] = record["id"]
            obj[STATUS_KEY] = record

        return objects

    def salesforce_collection_update(self, objects):
        """Updates records described as Robot/Python dictionaries.

        _objects_ is a dictionary of data in the format returned
        by the *Salesforce Collection Insert* keyword.

        A 200 record limit is enforced by the Salesforce APIs.

        Example:

        The following example creates ten accounts and then updates
        the Rating from "Cold" to "Hot"

        | ${data}=  Generate Test Data  Account  10
        | ...  Name=Account #{{number}}
        | ...  Rating=Cold
        | ${accounts}=  Salesforce Collection Insert  ${data}
        |
        | FOR  ${account}  IN  @{accounts}
        |     Set to dictionary  ${account}  Rating  Hot
        | END
        | Salesforce Collection Update  ${accounts}

        """
        for obj in objects:
            assert obj[
                "id"
            ], "Should be a list of objects with Ids returned by Salesforce Collection Insert"
            if STATUS_KEY in obj:
                del obj[STATUS_KEY]

        assert len(objects) <= SF_COLLECTION_INSERTION_LIMIT, (
            "Cannot update more than %s objects with this keyword"
            % SF_COLLECTION_INSERTION_LIMIT
        )

        records = self.cumulusci.sf.restful(
            "composite/sobjects",
            method="PATCH",
            json={"allOrNone": True, "records": objects},
        )

        for record, obj in zip(records, objects):
            obj[STATUS_KEY] = record

        for idx, (record, obj) in enumerate(zip(records, objects)):
            if record["errors"]:
                raise AssertionError(
                    "Error on Object {idx}: {record} : {obj}".format(**vars())
                )

    def salesforce_query(self, obj_name, **kwargs):
        """Constructs and runs a simple SOQL query and returns a list of dictionaries.

        By default the results will only contain object Ids. You can
        specify a SOQL SELECT clause via keyword arguments by passing
        a comma-separated list of fields with the ``select`` keyword
        argument.

        You can supply keys and values to match against
        in keyword arguments, or a full SOQL where-clause
        in a keyword argument named ``where``. If you supply
        both, they will be combined with a SOQL "AND".

        ``order_by`` and ``limit`` keyword arguments are also
        supported as shown below.

        Examples:

        The following example searches for all Contacts where the
        first name is "Eleanor". It returns the "Name" and "Id"
        fields and logs them to the robot report:

        | @{records}=  Salesforce Query  Contact  select=Id,Name
        | ...          FirstName=Eleanor
        | FOR  ${record}  IN  @{records}
        |     log  Name: ${record['Name']} Id: ${record['Id']}
        | END

        Or with a WHERE-clause, we can look for the last contact where
        the first name is NOT Eleanor.

        | @{records}=  Salesforce Query  Contact  select=Id,Name
        | ...          where=FirstName!='Eleanor'
        | ...              order_by=LastName desc
        | ...              limit=1
        """
        query = self._soql_query_builder(obj_name, **kwargs)
        self.builtin.log("Running SOQL Query: {}".format(query))
        return self.cumulusci.sf.query_all(query).get("records", [])

    def _soql_query_builder(
        self, obj_name, select=None, order_by=None, limit=None, where=None, **kwargs
    ):
        query = "SELECT "
        if select:
            query += select
        else:
            query += "Id"
        query += " FROM {}".format(obj_name)
        where_clauses = []
        if where:
            where_clauses = [where]
        for key, value in kwargs.items():
            where_clauses.append("{} = '{}'".format(key, value))
        if where_clauses:
            query += " WHERE " + " AND ".join(where_clauses)
        if order_by:
            query += " ORDER BY " + order_by
        if limit:
            assert int(limit), "Limit should be an integer"
            query += f" LIMIT {limit}"

        return query

    def salesforce_update(self, obj_name, obj_id, **kwargs):
        """Updates a Salesforce object by Id.

        The keyword returns the result from the underlying
        simple_salesforce ``insert`` method, which is an HTTP
        status code. As with `Salesforce Insert`, field values
        are specified as keyword arguments.

        The following example assumes that ${contact id} has been
        previously set, and adds a Description to the given
        contact.

        | &{contact}=  Salesforce Update  Contact  ${contact id}
        | ...  Description=This Contact created during a test
        | Should be equal as numbers ${result}  204

        """
        self.builtin.log(
            "Updating {} {} with values {}".format(obj_name, obj_id, kwargs)
        )
        obj_class = getattr(self.cumulusci.sf, obj_name)
        return obj_class.update(obj_id, kwargs)

    def soql_query(self, query, *args):
        """Runs a SOQL query and returns the result as a dictionary.

        The _query_ parameter must be a properly quoted SOQL query
        statement or statement fragment. Additional arguments will be
        joined to the query with spaces, allowing for a query to span
        multiple lines.

        This keyword will return a dictionary. The dictionary contains
        the keys as documented for the raw API call. The most useful
        keys are ``records`` and ``totalSize``, which contains a list
        of records that were matched by the query and the number of
        records that were returned.

        Example:

        The following example searches for all Contacts with a first
        name of "Eleanor" and a last name of "Rigby", and then logs
        the Id of the first record found.

        | ${result}=  SOQL Query
        | ...  SELECT Name, Id
        | ...  FROM   Contact
        | ...  WHERE  FirstName='Eleanor' AND LastName='Rigby'
        |
        | ${contact}=  Get from list  ${result['records']}  0
        | log  Contact Id: ${contact['Id']}

        """
        query = " ".join((query,) + args)
        self.builtin.log("Running SOQL Query: {}".format(query))
        return self.cumulusci.sf.query_all(query)

    def store_session_record(self, obj_type, obj_id):
        """Stores a Salesforce record's Id for use in the *Delete Session Records* keyword.

        This keyword is automatically called by *Salesforce Insert*.
        """
        self.builtin.log("Storing {} {} to session records".format(obj_type, obj_id))
        self._session_records.append({"type": obj_type, "id": obj_id})



import functools
import logging
import re
from typing import Optional, Tuple

from cumulusci.core.config import BaseProjectConfig, UniversalConfig
from cumulusci.core.exceptions import VcsException, VcsNotFoundError
from cumulusci.utils.yaml.cumulusci_yml import cci_safe_load
from cumulusci.vcs.base import VCSService
from cumulusci.vcs.models import (
    AbstractBranch,
    AbstractGitTag,
    AbstractPullRequest,
    AbstractRef,
    AbstractRelease,
    AbstractRepo,
    AbstractRepoCommit,
)


def get_service(
    config: BaseProjectConfig, logger: Optional[logging.Logger] = None
) -> VCSService:
    """Gets the VCS service based on the configuration.
    This function retrieves the VCS service type based on the CumulusCI
    configuration provided.
    If the service type is not specified in the configuration, it defaults to "github".
    The VCS Service class is expected to be defined in the configuration
    under the "class_path" key. If the class path is not found, a CumulusCIException is raised.

    Args:
        config (BaseProjectConfig): The configuration dictionary for the VCS service.

    Raises:
        CumulusCIException: Raised when the provider class is not found in the config.

    Returns:
        VCSService: The VCS service provider class.
    """
    vcs_service: VCSService = get_service_for_repo_url(config, config.repo_url)
    vcs_service.logger = logger or logging.getLogger(__name__)

    config.repo_service = vcs_service

    return vcs_service


def get_ref_for_tag(repo: AbstractRepo, tag_name: str) -> AbstractRef:
    """Gets a Reference object for the tag with the given name"""
    return repo.get_ref_for_tag(tag_name)


def get_tag_by_name(repo: AbstractRepo, tag_name: str) -> AbstractGitTag:
    """Fetches a tag by name from the given repository"""
    ref: AbstractRef = get_ref_for_tag(repo, tag_name)
    return repo.get_tag_by_ref(ref, tag_name)


VERSION_ID_RE: re.Pattern[str] = re.compile(r"version_id: (\S+)")


def get_version_id_from_commit(
    repo: AbstractRepo, commit_sha: str, context: str
) -> Optional[str]:
    """Fetches the version ID from a commit status"""
    commit: Optional[AbstractRepoCommit] = get_commit(repo, commit_sha)
    if commit is not None:
        version_id = commit.get_statuses(context, regex_match=VERSION_ID_RE)
        return version_id
    return None


def get_commit(repo: AbstractRepo, commit_sha: str) -> Optional[AbstractRepoCommit]:
    """Given a SHA1 hash, retrieve a AbstractRepoCommit object."""
    return repo.get_commit(commit_sha)


def get_pull_requests_with_base_branch(
    repo: AbstractRepo,
    base_branch_name: str,
    head: Optional[str] = None,
    state: Optional[str] = None,
) -> list:
    """Returns a list of pull requests with the given base branch"""
    if head:
        head = repo.owner_login + ":" + head
    return list(repo.pull_requests(base=base_branch_name, head=head, state=state))


def is_pull_request_merged(pull_request: AbstractPullRequest) -> bool:
    """Takes a AbstractPullRequest object"""
    return pull_request.merged_at is not None


def is_label_on_pull_request(
    repo: AbstractRepo, pull_request: AbstractPullRequest, label_name: str
) -> bool:
    """Returns True if the given label is on the pull request with the given
    pull request number. False otherwise."""
    labels = list(repo.get_pr_issue_labels(pull_request))

    return any(label_name == issue_label for issue_label in labels)


@functools.lru_cache(50)
def get_repo_from_url(
    config: BaseProjectConfig, url: str, service_alias: Optional[str] = None
) -> AbstractRepo:
    vcs_service: VCSService = get_service_for_repo_url(
        config, url, service_alias=service_alias
    )

    repo = vcs_service.get_repository(options={"repository_url": url})

    if repo is None:
        raise VcsNotFoundError(f"Could not find a repository at {url}.")

    return repo


@functools.lru_cache(50)
def get_service_for_repo_url(
    config: BaseProjectConfig, url: str, service_alias: Optional[str] = None
) -> VCSService:
    """Determines the VCS service type for the repository."""

    for service in VCSService.registered_services():
        try:
            vcs_service: Optional[VCSService] = service.get_service_for_url(
                config, url, service_alias=service_alias
            )

            if vcs_service is None:
                continue

            return vcs_service
        except NotImplementedError:
            # If the service does not implement get_service_for_url, skip it
            continue

    raise VcsException(
        f"Could not find a VCS service for URL: {url}. "
        "Please ensure the URL is correct and the service is properly configured."
    )


@functools.lru_cache(50)
def get_remote_project_config(repo: AbstractRepo, ref: str) -> BaseProjectConfig:
    contents_io = repo.file_contents("cumulusci.yml", ref=ref)
    return BaseProjectConfig(UniversalConfig(), cci_safe_load(contents_io))


def find_repo_feature_prefix(repo: AbstractRepo) -> str:
    ref = repo.branch(repo.default_branch).commit.sha
    head_cumulusci_project_config = get_remote_project_config(repo, ref)
    return head_cumulusci_project_config.project__git__prefix_feature or "feature/"


def get_remote_context(
    repo: AbstractRepo, commit_status_context: str, default_context: str
) -> str:
    config = get_remote_project_config(repo, repo.default_branch)
    return config.lookup(f"project__git__{commit_status_context}") or default_context


def find_latest_release(
    repo: AbstractRepo, include_beta=None
) -> Optional[AbstractRelease]:
    try:
        if include_beta:
            return get_latest_prerelease(repo)
        else:
            return repo.latest_release()
    except (VcsNotFoundError, StopIteration):
        pass


def get_latest_prerelease(repo: AbstractRepo) -> Optional[AbstractRelease]:
    """Calls GraphQL to retrieve the latest release, ordered chronologically."""
    return repo.get_latest_prerelease()


def find_previous_release(repo: AbstractRepo, prefix=None):
    most_recent = None
    for release in repo.releases():
        if prefix and not release.tag_name.startswith(prefix):
            continue
        if not prefix and release.prerelease:
            continue
        # Return the second release
        if most_recent is None:
            most_recent = release
        else:
            return release


def locate_commit_status_package_id(
    remote_repo: AbstractRepo, release_branch: AbstractBranch, context_2gp: str
) -> Tuple[Optional[str], Optional[AbstractRepoCommit]]:
    """Given a branch on a remote repo, walk the first 5 commits looking
    for a commit status equal to context_2gp and attempt to parse a
    package version id from the commit status detail."""
    version_id = None
    count = 0
    commit: Optional[AbstractRepoCommit] = release_branch.commit
    while version_id is None and count < 5:
        version_id = get_version_id_from_commit(remote_repo, commit.sha, context_2gp)
        if version_id:
            break
        count += 1
        if commit.parents:
            commit = remote_repo.get_commit(commit.parents[0].sha)
        else:
            commit = None
            break

    return version_id, commit


def get_repo_from_config(config: BaseProjectConfig, options: dict = {}) -> AbstractRepo:
    """Get a repository from the project config."""
    vcs_service: VCSService = get_service(config, logger=config.logger)

    return vcs_service.get_repository(options=options)


def get_latest_tag(repo: AbstractRepo, beta: bool = False) -> str:
    """Query Github Releases to find the latest production or beta tag"""
    prefix = repo.project_config.project__git__prefix_release

    try:
        if not beta:
            release: Optional[AbstractRelease] = repo.latest_release()

            if not release.tag_name.startswith(prefix):
                return _get_latest_tag_for_prefix(repo, prefix)

            return release.tag_name
        else:
            return _get_latest_tag_for_prefix(
                repo, repo.project_config.project__git__prefix_beta
            )
    except Exception:
        raise VcsException(
            f"No release found for {repo.repo_url} with tag prefix {prefix}"
        )


def _get_latest_tag_for_prefix(repo: AbstractRepo, prefix: str) -> str:
    for release in repo.releases():
        if not release.tag_name.startswith(prefix):
            continue
        return release.tag_name
    raise VcsException(f"No release found for {repo.repo_url} with tag prefix {prefix}")


def get_version_id_from_tag(repo: AbstractRepo, tag_name: str) -> str:
    """Given the name of a tag, return the version_id in the tag's message.

    @param tag_name: the name of the tag
    @param repo: the repository of the package to look for a release in
    @returns: the 04t id in the tag's messages
    """
    tag = get_tag_by_name(repo, tag_name)
    for line in tag.message.split("\n"):
        if line.startswith("version_id:"):
            version_id = line.split("version_id: ")[1]
            if not version_id.startswith("04t"):
                continue
            return version_id

    raise VcsException(f"Could not find version_id for tag {tag_name}")



from urllib.parse import urlparse

import simple_salesforce
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry

from cumulusci import __version__
from cumulusci.core.exceptions import ServiceNotConfigured, ServiceNotValid

CALL_OPTS_HEADER_KEY = "Sforce-Call-Options"


def get_simple_salesforce_connection(
    project_config, org_config, api_version=None, base_url: str = None
):

    # Retry on long-running metadeploy jobs
    retries = Retry(total=5, status_forcelist=(502, 503, 504), backoff_factor=0.3)
    adapter = HTTPAdapter(max_retries=retries)
    instance = org_config.instance_url

    # Attempt to get the host and port from the URL
    instance_url = urlparse(org_config.instance_url)
    instance = instance_url.hostname
    port = instance_url.port

    if port:
        instance = f"{instance}:{port}"

    sf = simple_salesforce.Salesforce(
        instance=instance,
        session_id=org_config.access_token,
        version=api_version or project_config.project__package__api_version,
    )
    try:
        app = project_config.keychain.get_service("connectedapp")
        client_name = app.client_id
    except (ServiceNotValid, ServiceNotConfigured):
        client_name = "CumulusCI/{}".format(__version__)

    sf.headers.setdefault(CALL_OPTS_HEADER_KEY, "client={}".format(client_name))
    sf.session.mount("http://", adapter)
    sf.session.mount("https://", adapter)

    if base_url:
        base_url = (
            base_url.strip("/") + "/"
        )  # exactly one training slash and no leading slashes
        sf.base_url += base_url

    return sf



from unittest import mock

from cumulusci.tasks.salesforce import DescribeMetadataTypes

from .util import create_task


class TestRetrieveMetadataTypes:
    def test_run_task(self):
        task = create_task(DescribeMetadataTypes)
        task._get_api = mock.Mock()
        task()
        task._get_api.assert_called_once()

    def test_run_task_with_apiversion(self):
        task = create_task(DescribeMetadataTypes, {"api_version": 8.0})
        assert task.options.get("api_version") == 8.0
        task._get_api()



import io
import os
import shutil
import tempfile
from datetime import datetime, timedelta
from pathlib import Path
from unittest import mock

import pytest
import yaml

from cumulusci.core.config import (
    BaseProjectConfig,
    ScratchOrgConfig,
    ServiceConfig,
    SfdxOrgConfig,
    UniversalConfig,
)
from cumulusci.core.exceptions import (
    NotInProject,
    ProjectConfigNotFound,
    ScratchOrgException,
    ServiceNotConfigured,
    SfdxOrgException,
)
from cumulusci.utils import cd, temporary_dir

__location__ = os.path.dirname(os.path.realpath(__file__))


@pytest.fixture
def scratch_def_file():
    with temporary_dir():
        with open("tmp.json", "w") as f:
            f.write("{}")
        yield


@mock.patch("pathlib.Path.home")
class TestUniversalConfig:
    def setup_method(self):
        self.tempdir_home = Path(tempfile.mkdtemp())

    def teardown_method(self):
        shutil.rmtree(self.tempdir_home)

    def _create_universal_config_local(self, content):
        global_config_dir = os.path.join(self.tempdir_home, ".cumulusci")
        os.makedirs(global_config_dir)
        filename = os.path.join(global_config_dir, UniversalConfig.config_filename)
        self._write_file(filename, content)

    def _write_file(self, filename, content):
        with open(filename, "w") as f:
            f.write(content)

    def test_load_universal_config_no_local(self, mock_class):
        mock_class.return_value = self.tempdir_home
        # clear cache
        UniversalConfig.config = None
        config = UniversalConfig()
        with open(__location__ + "/../../../cumulusci.yml", "r") as f_expected_config:
            expected_config = yaml.safe_load(f_expected_config)
        assert config.config == expected_config

    def test_load_universal_config_empty_local(self, mock_class):
        self._create_universal_config_local("")
        # clear cache
        UniversalConfig.config = None
        mock_class.return_value = self.tempdir_home
        config = UniversalConfig()
        with open(__location__ + "/../../../cumulusci.yml", "r") as f_expected_config:
            expected_config = yaml.safe_load(f_expected_config)
        assert config.config == expected_config

    def test_load_universal_config_with_local(self, mock_class):
        local_yaml = "tasks:\n    newtesttask:\n        description: test description"
        self._create_universal_config_local(local_yaml)
        mock_class.return_value = self.tempdir_home

        # clear cache
        UniversalConfig.config = None

        config = UniversalConfig()
        with open(__location__ + "/../../../cumulusci.yml", "r") as f_expected_config:
            expected_config = yaml.safe_load(f_expected_config)
        expected_config["tasks"]["newtesttask"] = {}
        expected_config["tasks"]["newtesttask"]["description"] = "test description"
        assert config.config == expected_config


@mock.patch("pathlib.Path.home")
class TestBaseProjectConfig:
    def setup_method(self):
        self.tempdir_home = Path(tempfile.mkdtemp())
        self.tempdir_project = tempfile.mkdtemp()
        self.project_name = "TestRepo"
        self.current_commit = "abcdefg1234567890"
        self.current_branch = "main"

    def teardown_method(self):
        shutil.rmtree(self.tempdir_home)
        shutil.rmtree(self.tempdir_project)

    def _create_git_config(self):

        filename = os.path.join(self.tempdir_project, ".git", "config")
        content = '[remote "origin"]\n' + "  url = git@github.com:TestOwner/{}".format(
            self.project_name
        )
        self._write_file(filename, content)

        filename = os.path.join(self.tempdir_project, ".git", "HEAD")
        content = "ref: refs/heads/{}".format(self.current_branch)
        self._write_file(filename, content)

        dirname = os.path.join(self.tempdir_project, ".git", "refs", "heads")
        os.makedirs(dirname)
        filename = os.path.join(dirname, "main")
        content = self.current_commit
        self._write_file(filename, content)

    def _create_universal_config_local(self, content):
        global_config_dir = os.path.join(self.tempdir_home, ".cumulusci")
        os.makedirs(global_config_dir)
        filename = os.path.join(global_config_dir, UniversalConfig.config_filename)
        self._write_file(filename, content)

    def _create_project_config(self):
        filename = os.path.join(self.tempdir_project, BaseProjectConfig.config_filename)
        content = (
            "project:\n"
            + "    name: TestRepo\n"
            + "    package:\n"
            + "        name: TestProject\n"
            + "        namespace: testproject\n"
        )
        self._write_file(filename, content)

    def _create_project_config_local(self, content):
        project_local_dir = os.path.join(
            self.tempdir_home, ".cumulusci", self.project_name
        )
        os.makedirs(project_local_dir)
        filename = os.path.join(project_local_dir, BaseProjectConfig.config_filename)
        self._write_file(filename, content)

    def _write_file(self, filename, content):
        with open(filename, "w") as f:
            f.write(content)

    def test_load_project_config_not_repo(self, mock_class):
        mock_class.return_value = self.tempdir_home
        with cd(self.tempdir_project):
            universal_config = UniversalConfig()
            with pytest.raises(NotInProject):
                BaseProjectConfig(universal_config)

    def test_load_project_config_no_config(self, mock_class):
        mock_class.return_value = self.tempdir_home
        os.mkdir(os.path.join(self.tempdir_project, ".git"))
        with cd(self.tempdir_project):
            universal_config = UniversalConfig()
            with pytest.raises(ProjectConfigNotFound):
                BaseProjectConfig(universal_config)

    def test_load_project_config_empty_config(self, mock_class):
        mock_class.return_value = self.tempdir_home
        os.mkdir(os.path.join(self.tempdir_project, ".git"))
        self._create_git_config()
        # create empty project config file
        filename = os.path.join(self.tempdir_project, BaseProjectConfig.config_filename)
        content = ""
        self._write_file(filename, content)

        with cd(self.tempdir_project):
            universal_config = UniversalConfig()
            config = BaseProjectConfig(universal_config)
            assert config.config_project == {}

    def test_load_project_config_valid_config(self, mock_class):
        mock_class.return_value = self.tempdir_home
        os.mkdir(os.path.join(self.tempdir_project, ".git"))
        self._create_git_config()
        local_yaml = "tasks:\n    newtesttask:\n        description: test description"
        self._create_universal_config_local(local_yaml)

        # create valid project config file
        self._create_project_config()

        with cd(self.tempdir_project):
            universal_config = UniversalConfig()
            config = BaseProjectConfig(universal_config)
            assert config.project__package__name == "TestProject"
            assert config.project__package__namespace == "testproject"

    def test_repo_owner(self, mock_class):
        mock_class.return_value = self.tempdir_home
        os.mkdir(os.path.join(self.tempdir_project, ".git"))
        self._create_git_config()

        # create valid project config file
        self._create_project_config()

        with cd(self.tempdir_project):
            universal_config = UniversalConfig()
            config = BaseProjectConfig(universal_config)
            assert config.repo_owner == "TestOwner"

    def test_repo_branch(self, mock_class):
        mock_class.return_value = self.tempdir_home
        os.mkdir(os.path.join(self.tempdir_project, ".git"))
        self._create_git_config()

        # create valid project config file
        self._create_project_config()

        with cd(self.tempdir_project):
            universal_config = UniversalConfig()
            config = BaseProjectConfig(universal_config)
            assert config.repo_branch == self.current_branch

    def test_repo_commit(self, mock_class):
        mock_class.return_value = self.tempdir_home
        os.mkdir(os.path.join(self.tempdir_project, ".git"))
        self._create_git_config()

        # create valid project config file
        self._create_project_config()

        with cd(self.tempdir_project):
            universal_config = UniversalConfig()
            config = BaseProjectConfig(universal_config)
            assert config.repo_commit == self.current_commit

    def test_load_project_config_local(self, mock_class):
        mock_class.return_value = self.tempdir_home
        os.mkdir(os.path.join(self.tempdir_project, ".git"))
        self._create_git_config()

        # create valid project config file
        self._create_project_config()

        # create local project config file
        content = "project:\n" + "    package:\n" + "        api_version: 45.0\n"
        self._create_project_config_local(content)

        with cd(self.tempdir_project):
            universal_config = UniversalConfig()
            config = BaseProjectConfig(universal_config)
            assert config.config_project_local != {}
            assert config.project__package__api_version == 45.0

    def test_load_additional_yaml(self, mock_class):
        mock_class.return_value = self.tempdir_home
        os.mkdir(os.path.join(self.tempdir_project, ".git"))
        self._create_git_config()

        # create valid project config file
        self._create_project_config()

        # create local project config file
        content = "project:\n" + "    package:\n" + "        api_version: 45.0\n"

        with cd(self.tempdir_project):
            universal_config = UniversalConfig()
            config = BaseProjectConfig(universal_config, additional_yaml=content)
            assert config.config_additional_yaml != {}
            assert config.project__package__api_version == 45.0


@mock.patch("sarge.Command")
class TestScratchOrgConfig:
    def test_scratch_info(self, Command):
        result = b"""{
    "result": {
        "instanceUrl": "url",
        "accessToken": "access!token",
        "username": "username",
        "password": "password",
        "createdDate": "1970-01-01T00:00:00Z",
        "expirationDate": "1970-01-08"
    }
}"""
        Command.return_value = mock.Mock(
            stderr=io.BytesIO(b""), stdout=io.BytesIO(result), returncode=0
        )

        config = ScratchOrgConfig({"username": "test", "created": True}, "test")
        info = config.scratch_info

        assert info == {
            "access_token": "access!token",
            "instance_url": "url",
            "org_id": "access",
            "password": "password",
            "username": "username",
            "created_date": "1970-01-01T00:00:00Z",
            "expiration_date": "1970-01-08",
        }
        assert info is config._sfdx_info
        for key in (
            "access_token",
            "instance_url",
            "org_id",
            "password",
            "username",
        ):
            assert key in config.config
        assert config._sfdx_info_date

    def test_scratch_info_memoized(self, Command):
        config = ScratchOrgConfig({"username": "test", "created": True}, "test")
        config._sfdx_info = _marker = object()
        info = config.scratch_info
        assert info is _marker

    def test_sfdx_info_non_json_response(self, Command):
        Command.return_value = mock.Mock(
            stderr=io.BytesIO(b""), stdout=io.BytesIO(b"<html></html>"), returncode=0
        )

        config = SfdxOrgConfig({"username": "test", "created": True}, "test")
        with pytest.raises(SfdxOrgException):
            config.sfdx_info

    def test_scratch_info_command_error(self, Command):
        Command.return_value = mock.Mock(
            stderr=io.BytesIO(b"error"), stdout=io.BytesIO(b"out"), returncode=1
        )

        config = ScratchOrgConfig(
            {"username": "test", "email_address": "test@example.com", "created": True},
            "test",
        )

        with pytest.raises(SfdxOrgException, match="error"):
            config.scratch_info

    def test_scratch_info_username_not_found(self, Command):
        Command.return_value = mock.Mock(
            stderr=io.BytesIO(b"error"), stdout=io.BytesIO(b"out"), returncode=0
        )

        config = ScratchOrgConfig(
            {"config_file": "tmp.json", "email_address": "test@example.com"}, "test"
        )

        with temporary_dir():
            with open("tmp.json", "w") as f:
                f.write("{}")
            with pytest.raises(ScratchOrgException):
                config.scratch_info

    def test_access_token(self, Command):
        config = ScratchOrgConfig({}, "test")
        _marker = object()
        config._sfdx_info = {"access_token": _marker}
        assert config.access_token is _marker

    def test_get_access_token(self, Command):
        """Verify that get_access_token calls out to sfdx"""
        sf = mock.Mock()
        sf.query_all.return_value = {"records": [{"Username": "whatever@example.com"}]}

        sfdx_response = mock.Mock(returncode=0)
        sfdx_response.stdout_text.read.return_value = (
            '{"result": {"accessToken": "the-token"}}'
        )
        sfdx = mock.Mock(return_value=sfdx_response)

        config = ScratchOrgConfig({}, "test")
        with mock.patch(
            "cumulusci.core.config.org_config.OrgConfig.salesforce_client", sf
        ):
            with mock.patch("cumulusci.core.config.sfdx_org_config.sfdx", sfdx):
                access_token = config.get_access_token(alias="dadvisor")
                sfdx.assert_called_once_with(
                    "org display --target-org=whatever@example.com --json"
                )
                assert access_token == "the-token"

    def test_get_access_token__default(self, Command):
        """Verify that with no args, get_access_token returns the default token"""
        config = ScratchOrgConfig({}, "test")
        _marker = object()
        config._sfdx_info = {"access_token": _marker}
        access_token = config.get_access_token()
        assert access_token is _marker

    def test_get_access_token__unknown_user(self, Command):
        sf = mock.Mock()
        sf.query_all.return_value = {"records": []}

        config = ScratchOrgConfig({}, "test")

        with mock.patch(
            "cumulusci.core.config.org_config.OrgConfig.salesforce_client", sf
        ):
            with pytest.raises(
                SfdxOrgException,
                match="Couldn't find a username for the specified user",
            ):
                config.get_access_token(alias="dadvisor")

    def test_get_access_token__multiple_users(self, Command):
        sf = mock.Mock()
        sf.query_all.return_value = {
            "records": [
                {"Username": "test1@example.com"},
                {"Username": "test2@example.com"},
            ]
        }

        config = ScratchOrgConfig({}, "test")

        with mock.patch(
            "cumulusci.core.config.org_config.OrgConfig.salesforce_client", sf
        ):
            with pytest.raises(
                SfdxOrgException,
                match="More than one user matched the search critiera.",
            ):
                config.get_access_token(alias="dadvisor")

    def test_get_access_token__no_access_token(self, Command):
        sf = mock.Mock()
        sf.query_all.return_value = {"records": [{"Username": "whatever@example.com"}]}

        sfdx_response = mock.Mock(returncode=1)
        sfdx_response.stdout_text.read.return_value = '{"message": "blah blah..."}'
        sfdx = mock.Mock(return_value=sfdx_response)

        config = ScratchOrgConfig({}, "test")
        with mock.patch(
            "cumulusci.core.config.org_config.OrgConfig.salesforce_client", sf
        ):
            with mock.patch("cumulusci.core.config.sfdx_org_config.sfdx", sfdx):
                exception = (
                    "Unable to find access token for whatever@example.com\nblah blah..."
                )
                with pytest.raises(SfdxOrgException, match=exception):
                    config.get_access_token(alias="dadvisor")

    def test_instance_url(self, Command):
        config = ScratchOrgConfig({}, "test")
        _marker = object()
        config._sfdx_info = {"instance_url": _marker}
        assert config.instance_url is _marker

    def test_org_id_from_config(self, Command):
        config = ScratchOrgConfig({"org_id": "test"}, "test")
        assert config.org_id == "test"

    def test_org_id_from_sfdx_info(self, Command):
        config = ScratchOrgConfig({}, "test")
        _marker = object()
        config._sfdx_info = {"org_id": _marker}
        assert config.org_id is _marker

    def test_user_id_from_config(self, Command):
        config = ScratchOrgConfig({"user_id": "test"}, "test")
        assert config.user_id == "test"

    def test_user_id_from_org(self, Command):
        sf = mock.Mock()
        sf.query_all.return_value = {"records": [{"Id": "test"}]}

        config = ScratchOrgConfig({"username": "test_username"}, "test")
        config._sfdx_info = {
            "instance_url": "test_instance",
            "access_token": "token",
        }
        with mock.patch(
            "cumulusci.core.config.org_config.OrgConfig.salesforce_client", sf
        ):
            assert config.user_id == "test"

    def test_username_from_sfdx_info(self, Command):
        config = ScratchOrgConfig({}, "test")
        _marker = object()
        config._sfdx_info = {"username": _marker}
        assert config.username is _marker

    def test_password_from_config(self, Command):
        config = ScratchOrgConfig({"password": "test"}, "test")
        assert config.password == "test"

    def test_password_from_sfdx_info(self, Command):
        config = ScratchOrgConfig({}, "test")
        _marker = object()
        config._sfdx_info = {"password": _marker}
        assert config.password is _marker

    def test_email_address_from_config(self, Command):
        config = ScratchOrgConfig({"email_address": "test@example.com"}, "test")

        assert config.email_address == "test@example.com"
        Command.return_value.assert_not_called()

    def test_email_address_from_git(self, Command):
        config = ScratchOrgConfig({}, "test")
        Command.return_value = p = mock.Mock(
            stdout=io.BytesIO(b"test@example.com"), stderr=io.BytesIO(b""), returncode=0
        )

        assert config.email_address == "test@example.com"
        config.email_address  # Make sure value is cached
        p.run.assert_called_once()

    def test_email_address_not_present(self, Command):
        config = ScratchOrgConfig({}, "test")
        Command.return_value = p = mock.Mock(
            stdout=io.BytesIO(b""), stderr=io.BytesIO(b""), returncode=0
        )

        assert config.email_address is None
        p.run.assert_called_once()

    def test_days(self, Command):
        config = ScratchOrgConfig({"days": 2}, "test")
        assert config.days == 2

    def test_days_default(self, Command):
        config = ScratchOrgConfig({}, "test")
        assert config.days == 1

    def test_format_days(self, Command):
        config = ScratchOrgConfig({"days": 2}, "test")
        assert config.format_org_days() == "2"
        now = datetime.now()
        config.date_created = now
        assert config.format_org_days() == "1/2"
        config.date_created = now - timedelta(days=3)
        assert config.format_org_days() == "2"

    def test_expired(self, Command):
        config = ScratchOrgConfig({"days": 1}, "test")
        now = datetime.now()
        config.date_created = now
        assert not config.expired
        config.date_created = now - timedelta(days=2)
        assert config.expired

    def test_expires(self, Command):
        config = ScratchOrgConfig({"days": 1}, "test")
        now = datetime.now()
        config.date_created = now
        assert config.expires == now + timedelta(days=1)

        config = ScratchOrgConfig({"days": 1}, "test")
        config.date_created = None
        assert config.expires is None

    def test_days_alive(self, Command):
        config = ScratchOrgConfig({}, "test")
        config.date_created = datetime.now()
        assert config.days_alive == 1

    def test_active(self, Command):
        config = ScratchOrgConfig({}, "test")
        config.date_created = None
        assert not config.active

        config = ScratchOrgConfig({}, "test")
        config.date_created = datetime.now()
        assert config.active

        config = ScratchOrgConfig({}, "test")
        config.date_created = datetime.now() - timedelta(days=10)
        assert not config.active

    def test_create_org(self, Command):
        out = b"""{
                "status": 0,
                "result": {
                    "orgId": "ORG_ID",
                    "username": "USERNAME"
                }
                }
                """
        Command.return_value = p = mock.Mock(
            stdout=io.BytesIO(out), stderr=io.BytesIO(b""), returncode=0
        )

        config = ScratchOrgConfig(
            {
                "config_file": "tmp.json",
                "set_password": True,
                "email_address": "test@example.com",
            },
            "test",
        )
        config.generate_password = mock.Mock()
        with temporary_dir():
            with open("tmp.json", "w") as f:
                f.write("{}")

            config.create_org()

        p.run.assert_called_once()
        assert config.config["org_id"] == "ORG_ID"
        assert config.config["username"] == "USERNAME"
        assert "date_created" in config.config
        config.generate_password.assert_called_once()
        assert config.config["created"]
        assert config.scratch_org_type == "workspace"

    def test_check_apiversion_error(self, Command):
        out = b"""{
            "context": "Create",
            "commandName": "Create",
            "message": "The requested resource does not exist",
            "name": "NOT_FOUND"
            }"""

        Command.return_value = mock.Mock(
            stdout=io.BytesIO(out), stderr=io.BytesIO(b""), returncode=1
        )
        config = ScratchOrgConfig(
            {"config_file": "tmp.json", "email_address": "test@example.com"}, "test"
        )
        with temporary_dir():
            with open("tmp.json", "w") as f:
                f.write("{}")
            with pytest.raises(
                SfdxOrgException,
                match="The Salesforce CLI was unable to create a scratch org. Ensure you are connected using a valid API version on an active Dev Hub.",
            ):
                config.create_org()

    def test_create_org_no_config_file(self, Command):
        config = ScratchOrgConfig({}, "test")
        with pytest.raises(ScratchOrgException, match="missing a config_file"):
            config.create_org()
        Command.assert_not_called()

    def test_create_org_command_error(self, Command):
        Command.return_value = mock.Mock(
            stdout=io.BytesIO(b""), stderr=io.BytesIO(b"scratcherror"), returncode=1
        )

        config = ScratchOrgConfig(
            {"config_file": "tmp.json", "email_address": "test@example.com"}, "test"
        )
        with temporary_dir():
            with open("tmp.json", "w") as f:
                f.write("{}")

            with pytest.raises(ScratchOrgException) as e:
                config.create_org()
            assert "scratcherror" in str(e.value)

    def test_generate_password(self, Command):
        p = mock.Mock(
            stderr=io.BytesIO(b"error"), stdout=io.BytesIO(b"out"), returncode=0
        )
        Command.return_value = p

        config = ScratchOrgConfig({"username": "test"}, "test")
        config.generate_password()

        p.run.assert_called_once()

    def test_generate_password_failed(self, Command):
        p = mock.Mock()
        p.stderr = io.BytesIO(b"error")
        p.stdout = io.BytesIO(b"out")
        p.returncode = 1
        Command.return_value = p

        config = ScratchOrgConfig({"username": "test"}, "test")
        config.logger = mock.Mock()
        config.generate_password()

        config.logger.warning.assert_called_once()

    def test_generate_password_skips_if_failed(self, Command):
        config = ScratchOrgConfig({"username": "test"}, "test")
        config.password_failed = True
        config.generate_password()
        Command.assert_not_called()

    def test_can_delete(self, Command):
        config = ScratchOrgConfig({"date_created": datetime.now()}, "test")
        assert config.can_delete()

    def test_delete_org(self, Command):
        Command.return_value = mock.Mock(
            stdout=io.BytesIO(b"info"), stderr=io.BytesIO(b""), returncode=0
        )

        config = ScratchOrgConfig(
            {"username": "test", "created": True, "instance_url": "https://blah"},
            "test",
        )
        config.keychain = mock.Mock()
        config.delete_org()

        assert not config.config.get("instance_url")
        assert not config.config["created"]
        assert config.config["username"] is None

    def test_delete_org_not_created(self, Command):
        config = ScratchOrgConfig({"created": False}, "test")
        config.delete_org()
        Command.assert_not_called()

    def test_delete_org_error(self, Command):
        Command.return_value = mock.Mock(
            stdout=io.BytesIO(b"An error occurred deleting this org"),
            stderr=io.BytesIO(b""),
            returncode=1,
        )

        config = ScratchOrgConfig({"username": "test", "created": True}, "test")
        with pytest.raises(ScratchOrgException):
            config.delete_org()

    def test_force_refresh_oauth_token(self, Command):
        Command.return_value = p = mock.Mock(
            stdout=io.BytesIO(b""), stderr=io.BytesIO(b""), returncode=0
        )

        config = ScratchOrgConfig({"username": "test"}, "test")
        config.force_refresh_oauth_token()

        p.run.assert_called_once()

    def test_force_refresh_oauth_token_error(self, Command):
        Command.return_value = mock.Mock(
            stdout=io.BytesIO(b"error"), stderr=io.BytesIO(b""), returncode=1
        )

        config = ScratchOrgConfig({"username": "test"}, "test")
        with pytest.raises(SfdxOrgException):
            config.force_refresh_oauth_token()

    def test_refresh_oauth_token(self, Command):
        result = b"""{
    "result": {
        "instanceUrl": "url",
        "accessToken": "access!token",
        "username": "username",
        "password": "password",
        "createdDate": "1970-01-01T:00:00:00Z",
        "expirationDate": "1970-01-08"
    }
}"""
        Command.return_value = mock.Mock(
            stdout=io.BytesIO(result), stderr=io.BytesIO(b""), returncode=0
        )

        config = ScratchOrgConfig({"username": "test", "created": True}, "test")
        config._sfdx_info = {}
        config._sfdx_info_date = datetime.now() - timedelta(days=1)
        config.force_refresh_oauth_token = mock.Mock()
        config._load_orginfo = mock.Mock()

        config.refresh_oauth_token(keychain=None)

        config.force_refresh_oauth_token.assert_called_once()
        assert config._sfdx_info

    def test_choose_devhub(self, Command):
        mock_keychain = mock.Mock()
        mock_keychain.get_service.return_value = ServiceConfig(
            {"username": "fake@fake.devhub"}
        )
        config = ScratchOrgConfig({}, "test", mock_keychain)

        assert config._choose_devhub_username() == "fake@fake.devhub"

    def test_choose_devhub__service_not_configured(self, Command):
        mock_keychain = mock.Mock()
        mock_keychain.get_service.side_effect = ServiceNotConfigured
        config = ScratchOrgConfig({}, "test", mock_keychain)

        assert config._choose_devhub_username() is None


class TestScratchOrgConfigPytest:
    def test_build_org_create_args(self, scratch_def_file):
        mock_keychain = mock.Mock()
        mock_keychain.get_service.return_value = ServiceConfig(
            {"username": "fake@fake.devhub"}
        )
        config = ScratchOrgConfig(
            {
                "config_file": "tmp.json",
                "email_address": "test@example.com",
                "noancestors": True,
                "sfdx_alias": "project__org",
                "default": True,
                "release": "previous",
            },
            "test",
            mock_keychain,
        )
        args = config._build_org_create_args()
        assert args == [
            "-f",
            "tmp.json",
            "-w",
            "120",
            "--target-dev-hub",
            "fake@fake.devhub",
            "--no-namespace",
            "--no-ancestors",
            "--duration-days",
            "1",
            "--release=previous",
            "-a",
            "project__org",
            "--admin-email=test@example.com",
            "--set-default",
        ]

    def test_build_org_create_args__email_in_scratch_def(self):
        config = ScratchOrgConfig(
            {
                "config_file": "tmp.json",
                "set_password": True,
                "email_address": "test@example.com",
            },
            "test",
        )
        with temporary_dir():
            with open("tmp.json", "w") as f:
                f.write('{"adminEmail": "other_test@example.com"}')
            args = config._build_org_create_args()

        assert "adminEmail=test@example.com" not in args

    def test_temporary_backwards_compatiblity_hacks(self):
        filename = Path(__file__).parent / "_test_config_backwards_compatibility.py"
        # run this in a separate process to not confuse
        # the module table
        assert os.system(f"pytest {filename}") == 0



import pytest
import vcr

from cumulusci.utils.salesforce.count_sobjects import count_sobjects


class TestCountSObjects:
    @pytest.mark.vcr()
    def test_count_sobjects_simple(self, sf):
        results, net_errors, sf_errors = count_sobjects(sf, ["Account", "Opportunity"])
        assert "Account" in results
        assert "Opportunity" in results
        assert not net_errors and not sf_errors

    @pytest.mark.vcr()
    def test_count_sobjects__errors(self, sf):
        results, net_errors, sf_errors = count_sobjects(sf, ["Account", "XYZZY"])
        assert "Account" in results
        assert "XYZZY" not in results
        assert not net_errors
        assert sf_errors and "XYZZY" in str(sf_errors)

    @pytest.mark.vcr()
    def test_count_sobjects__network_errors(self, sf):
        network_errors_yaml = "cumulusci/utils/salesforce/tests/cassettes/ManualEdit_TestCountSObjects.test_count_sobjects__network_errors.yaml"
        with vcr.use_cassette(network_errors_yaml):
            _, net_errors, sf_errors = count_sobjects(sf, ["Account", "XYZZY"])
            assert net_errors
            assert not sf_errors



from concurrent.futures import ThreadPoolExecutor
from typing import Callable, Dict


class RunParallelQueries:
    @staticmethod
    def _run_queries_in_parallel(
        queries: Dict[str, str], run_query: Callable[[str], dict], num_threads: int = 4
    ) -> Dict[str, list]:
        """Accepts a set of queries structured as {'query_name': 'query'}
        and a run_query function that runs a particular query. Runs queries in parallel and returns the queries"""
        results_dict = {}

        with ThreadPoolExecutor(max_workers=num_threads) as executor:
            futures = {
                query_name: executor.submit(run_query, query)
                for query_name, query in queries.items()
            }

        for query_name, future in futures.items():
            try:
                query_result = future.result()
                results_dict[query_name] = query_result["records"]
            except Exception as e:
                raise Exception(f"Error executing query '{query_name}': {type(e)}: {e}")
            else:
                queries.pop(query_name, None)

        return results_dict



import os
from collections import defaultdict

from packaging import version

from cumulusci.core.exceptions import CumulusCIException, TaskOptionsError
from cumulusci.core.utils import process_bool_arg, process_list_arg
from cumulusci.tasks.metadata_etl import (
    MetadataOperation,
    MetadataSingleEntityTransformTask,
)
from cumulusci.tasks.salesforce import BaseSalesforceApiTask
from cumulusci.utils import CUMULUSCI_PATH
from cumulusci.utils.xml import metadata_tree


class ProfileGrantAllAccess(MetadataSingleEntityTransformTask, BaseSalesforceApiTask):
    name = "ProfileGrantAllAccess"
    entity = "Profile"

    task_options = {
        "package_xml": {
            "description": "Override the default package.xml file for retrieving the Admin.profile and all objects and classes "
            "that need to be included by providing a path to your custom package.xml"
        },
        "record_types": {
            "description": "A list of dictionaries containing the required key `record_type` with a value specifying "
            "the record type in format <object>.<developer_name>.  Record type names can use the token strings {managed} "
            "and {namespaced_org} for namespace prefix injection as needed.  By default, all listed record types will be set "
            "to visible and not default.  Use the additional keys `visible`, `default`, and `person_account_default` set to "
            "true/false to override.  "
            "Page Layout Support: If you are using the Page Layouts feature, you can specify the `page_layout` key with the "
            "layout name to use for the record type.  If not specified, the default page layout will be used.  "
            "NOTE: Setting record_types is only supported in cumulusci.yml, command line override is not supported."
        },
        "managed": {
            "description": "If True, uses the namespace prefix where appropriate.  Use if running against an org with the managed package "
            "installed.  Defaults to False"
        },
        "namespaced_org": {
            "description": "If True, attempts to prefix all unmanaged metadata references with the namespace prefix for deployment to the "
            "packaging org or a namespaced scratch org.  Defaults to False"
        },
        "namespace_inject": {
            "description": "If set, the namespace tokens in files and filenames are replaced with the namespace's prefix. "
            "Defaults to project__package__namespace"
        },
        "profile_name": {
            "description": "Name of the Profile to target for updates (deprecated; use api_names to target multiple profiles).",
            "default": "Admin",
        },
        "include_packaged_objects": {
            "description": "Automatically include objects from all installed managed packages. "
            "Defaults to True in projects that require CumulusCI 3.9.0 and greater that don't use a custom package.xml, otherwise False."
        },
        "api_names": {"description": "List of API names of Profiles to affect"},
    }

    def _init_options(self, kwargs):
        super(ProfileGrantAllAccess, self)._init_options(kwargs)

        self.options["namespace_inject"] = self.options.get(
            "namespace_inject", self.project_config.project__package__namespace
        )

        # We enable new functionality to extend the package.xml to packaged objects
        # by default only if we meet specific requirements: the project has to require
        # CumulusCI >= 3.9.0 (i.e., creation date or opt-in after release), and it must
        # not be using a custom `package.xml`
        min_cci_version = self.project_config.minimum_cumulusci_version
        if min_cci_version and "package_xml" not in self.options:
            parsed_version = version.parse(min_cci_version)
            default_packages_arg = parsed_version >= version.parse("3.9.0")
        else:
            default_packages_arg = False

        self.options["include_packaged_objects"] = process_bool_arg(
            self.options.get("include_packaged_objects", default_packages_arg)
        )

        # Build the api_names list, taking into account legacy behavior.
        # If we're using a custom package.xml, we will union the api_names list with
        # any Profiles specified there.
        self.api_names = set(process_list_arg(self.options.get("api_names") or []))
        if "profile_name" in self.options:
            self.api_names.add(self.options["profile_name"])
        if not self.api_names and "package_xml" not in self.options:
            self.api_names.add(
                "Admin"
            )  # Don't add a default if using custom package.xml

        if "package_xml" in self.options:
            self.package_xml_path = self.options["package_xml"]
        else:
            self.package_xml_path = os.path.join(
                CUMULUSCI_PATH, "cumulusci", "files", "admin_profile.xml"
            )

        if self.org_config is not None:
            # Set up namespace prefix strings.
            # We can only do this if we actually have an org_config;
            # i.e. not while freezing steps for metadeploy
            namespace = self.options["namespace_inject"]
            namespace_prefix = f"{namespace}__" if namespace else ""
            self.namespace_prefixes = {
                "managed": namespace_prefix if self.options["managed"] else "",
                "namespaced_org": namespace_prefix
                if self.options["namespaced_org"]
                else "",
            }
            self.api_names = {self._inject_namespace(x) for x in self.api_names}

    def freeze(self, step):
        # Preserve behavior from when we subclassed Deploy.

        steps = super().freeze(step)
        for step in steps:
            if step["kind"] == "other":
                step["kind"] = "metadata"
        return steps

    def _generate_package_xml(self, operation):
        if operation is MetadataOperation.RETRIEVE:
            with open(self.package_xml_path, "r", encoding="utf-8") as f:
                package_xml_content = f.read()

            package_xml_content = package_xml_content.format(**self.namespace_prefixes)

            # We need to rewrite the package.xml for one or two reasons.
            # Either we are using packaged-object expansion, or we're using
            # a package.xml and need to substitute in profile API names.

            # Convert to bytes because stored `package.xml`s typically have an encoding declaration,
            # which `fromstring()` doesn't like.
            package_xml = metadata_tree.fromstring(package_xml_content.encode("utf-8"))

            if self.options["include_packaged_objects"]:
                self._expand_package_xml(package_xml)

            self._expand_profile_members(package_xml)

            package_xml_content = package_xml.tostring(xml_declaration=True)

            return package_xml_content
        else:
            return super()._generate_package_xml(operation)

    def _expand_profile_members(self, package_xml):
        profile_names = package_xml.find("types", name="Profile")
        if not profile_names:
            profile_names = package_xml.append("types")
            profile_names.append("name", "Profile")

        listed_api_names = {p.text for p in profile_names.findall("members")}

        for profile in self.api_names:
            if profile not in listed_api_names:
                profile_names.append("members", text=profile)

        self.api_names.update(listed_api_names)

    def _expand_package_xml(self, package_xml):
        # Query the target org for all namespaced objects
        # Add these entities to the package.xml

        results = self.tooling.query_all(
            "SELECT DeveloperName, NamespacePrefix FROM CustomObject WHERE ManageableState != 'unmanaged'"
        )

        custom_objects = package_xml.find("types", name="CustomObject")
        if not custom_objects:
            raise CumulusCIException(
                "Unable to add packaged objects to package.xml because it does not contain a <types> tag of type CustomObject."
            )

        for record in results.get("records", []):
            custom_objects.append(
                "members",
                text=f"{record['NamespacePrefix']}__{record['DeveloperName']}__c",
            )

        self._expand_package_xml_objects(package_xml)

    def _expand_package_xml_objects(self, package_xml):
        # Check for any record types specified in the options, but missing from the package.xml
        # Add these entities to the package.xml

        custom_objects = package_xml.find("types", name="CustomObject")

        # Append custom objects if record types are present but missing from package.xml
        record_types = self.options.get("record_types") or []
        rt_objects = {rt["record_type"].split(".")[0] for rt in record_types}
        listed_custom_objects = {c.text for c in custom_objects.findall("members")}

        for rt in rt_objects:
            if rt not in listed_custom_objects:
                self.logger.info('Adding "{}" to package.xml'.format(rt))
                custom_objects.append(
                    "members",
                    text=rt,
                )

    def _transform_entity(self, tree, api_name):
        # Custom applications
        self._set_elements_visible(tree, "applicationVisibilities", "visible")
        # Apex classes
        self._set_elements_visible(tree, "classAccesses", "enabled")
        # Fields
        self._set_elements_visible(tree, "fieldPermissions", "editable")
        self._set_elements_visible(tree, "fieldPermissions", "readable")
        # Visualforce pages
        self._set_elements_visible(tree, "pageAccesses", "enabled")
        # Custom tabs
        self._set_elements_visible(
            tree,
            "tabVisibilities",
            "visibility",
            false_value="Hidden",
            true_value="DefaultOn",
        )
        # Record Types
        self._set_record_types(tree, api_name)

        return tree

    def _set_elements_visible(
        self, tree, outer_tag, inner_tag, false_value="false", true_value="true"
    ):
        for elem in tree.findall(outer_tag, **{inner_tag: false_value}):
            elem.find(inner_tag).text = true_value

    def _strip_namespace_from_record_type(self, record_type):
        """Strip namespace prefix from record type string.
        
        Converts 'namespace__Object__c.namespace__RecordType' to 'Object__c.RecordType'
        """
        if not self.options.get("namespace_inject"):
            return record_type
            
        namespace = self.options["namespace_inject"]
        namespace_prefix = f"{namespace}__"
        
        # Split on the dot to handle object and record type separately
        parts = record_type.split(".", 1)
        if len(parts) != 2:
            return record_type
            
        object_name, record_type_name = parts
        
        # Strip namespace prefix from both object and record type if present
        if object_name.startswith(namespace_prefix):
            object_name = object_name[len(namespace_prefix):]
        if record_type_name.startswith(namespace_prefix):
            record_type_name = record_type_name[len(namespace_prefix):]
            
        return f"{object_name}.{record_type_name}"

    def _set_record_types(self, tree, api_name):
        # Do namespace injection
        record_types = self.options.get("record_types") or []
        for rt in record_types:
            rt["record_type"] = rt["record_type"].format(**self.namespace_prefixes)

        # If defaults are specified, clear any pre-existing defaults
        # that apply to the same object
        defaults = {
            "default": "default",
            "person_account_default": "personAccountDefault",
        }
        objects_with_defaults = defaultdict(set)

        for option, default_element in defaults.items():
            for rt in record_types:
                if option in rt:
                    objects_with_defaults[option].add(rt["record_type"].split(".")[0])

            for elem in tree.findall("recordTypeVisibilities"):
                if (
                    elem.find(default_element)
                    and elem.recordType.text.split(".")[0]
                    in objects_with_defaults[option]
                ):
                    elem.find(default_element).text = "false"

        # Set recordTypeVisibilities
        for rt in record_types:
            # Look for the recordTypeVisibilities element
            # First try with the original record type
            elem = tree.find("recordTypeVisibilities", recordType=rt["record_type"])
            
            # If not found and we're in a namespaced org, try with namespace stripped 
            # (since Salesforce may return without namespace in namespaced orgs)
            if elem is None and self.options.get("namespaced_org"):
                record_type_without_namespace = self._strip_namespace_from_record_type(rt["record_type"])
                elem = tree.find("recordTypeVisibilities", recordType=record_type_without_namespace)
                
                if elem is None:
                    raise TaskOptionsError(
                        f"Record Type {rt['record_type']} (or {record_type_without_namespace}) not found in retrieved {api_name}.profile"
                    )
            elif elem is None:
                raise TaskOptionsError(
                    f"Record Type {rt['record_type']} not found in retrieved {api_name}.profile"
                )

            # Set visible
            elem.visible.text = str(rt.get("visible", "true")).lower()

            # Set default
            elem.default.text = str(rt.get("default", "false")).lower()

            # Set person account default if element exists
            pa_default = elem.find("personAccountDefault")
            if pa_default is not None:
                pa_default.text = str(rt.get("person_account_default", "false")).lower()

        # Set page layout defaults for record types
        for rt in record_types:
            # We need it to look like this:
            # <layoutAssignments>
            #   <layout>{page_layout}</layout>
            #   <recordType>{record_type}</recordType>
            # </layoutAssignments>
            layout_option = rt.get("page_layout", None)
            if layout_option:
                # Look for page layout definitions in the record type
                found_layout = False
                for elem in tree.findall("layoutAssignments"):
                    if elem.find("recordType").text == rt["record_type"]:
                        elem.layout.text = layout_option
                        found_layout = True

                if not found_layout:
                    assignment = tree.append(tag="layoutAssignments")
                    assignment.append(tag="recordType", text=rt["record_type"])
                    assignment.append(tag="layout", text=layout_option)


UpdateAdminProfile = UpdateProfile = ProfileGrantAllAccess



# pyright: strict

import typing as T

from cumulusci.salesforce_api.org_schema import Schema
from cumulusci.tasks.bulkdata.generate_mapping_utils.generate_mapping_from_declarations import (
    SimplifiedExtractDeclarationWithLookups,
    classify_and_filter_lookups,
)

from ..extract_dataset_utils.synthesize_extract_declarations import (
    ExtractDeclaration,
    flatten_declarations,
)


def _mapping_decl_for_extract_decl(
    decl: SimplifiedExtractDeclarationWithLookups,
):
    """Make a CCI extract mapping step from a SimplifiedExtractDeclarationWithLookups"""
    lookups = {lookup: {"table": tables} for lookup, tables in decl.lookups.items()}
    mapping_dict: dict[str, T.Any] = {
        "sf_object": decl.sf_object,
    }
    if decl.where:
        mapping_dict["soql_filter"] = decl.where
    if decl.api:
        mapping_dict["api"] = decl.api.value
    mapping_dict["fields"] = decl.fields
    if lookups:
        mapping_dict["lookups"] = lookups

    return (f"Extract {decl.sf_object}", mapping_dict)


def create_extract_mapping_file_from_declarations(
    decls: T.List[ExtractDeclaration], schema: Schema, opt_in_only: T.Sequence[str]
):
    """Create a mapping file sufficient for driving an extract process
    from an extract declarations file."""
    assert decls is not None
    simplified_decls = flatten_declarations(decls, schema, opt_in_only)
    simplified_decls = classify_and_filter_lookups(simplified_decls, schema)
    mappings = [_mapping_decl_for_extract_decl(decl) for decl in simplified_decls]
    return dict(pair for pair in mappings if pair)
